import segyio
from . import TraceSortingFormat
from . import SegySampleFormat

import numpy as np
import textwrap


def dt(f, fallback_dt=4000.0):
    """Delta-time

    Infer a ``dt``, the sample rate, from the file. If none is found, use the
    fallback.

    Parameters
    ----------

    f : segyio.SegyFile
    fallback_dt : float
        delta-time to fall back to, in microseconds

    Returns
    -------
    dt : float

    Notes
    -----

    .. versionadded:: 1.1

    """
    return f.xfd.getdt(fallback_dt)


def sample_indexes(segyfile, t0=0.0, dt_override=None):
    """
    Creates a list of values representing the samples in a trace at depth or time.
    The list starts at *t0* and is incremented with am*dt* for the number of samples.
    If a *dt_override* is not provided it will try to find a *dt* in the file.


    Parameters
    ----------
    segyfile :  segyio.SegyFile
    t0 : float
        initial sample, or delay-recording-time
    dt_override : float or None

    Returns
    -------
    samples : array_like of float

    Notes
    -----

    .. versionadded:: 1.1

    """
    if dt_override is None:
        dt_override = dt(segyfile)

    return [t0 + t * dt_override for t in range(len(segyfile.samples))]


def create_text_header(lines):
    """Format textual header

    Create a "correct" SEG-Y textual header.  Every line will be prefixed with
    C## and there are 40 lines. The input must be a dictionary with the line
    number[1-40] as a key. The value for each key should be up to 76 character
    long string.

    Parameters
    ----------

    lines : dict
        `lines` dictionary with fields:

        - ``no`` : line number (`int`)
        - ``line`` : line (`str`)

    Returns
    -------

    text : str

    """

    rows = []
    for line_no in range(1, 41):
        line = ""
        if line_no in lines:
            line = lines[line_no]
        row = "C{0:>2} {1:76}".format(line_no, line)
        rows.append(row)

    rows = ''.join(rows)
    return rows

def wrap(s, width=80):
    """
    Formats the text input with newlines given the user specified width for
    each line. `wrap` will attempt to decode the input as ascii, ignoring any
    errors, which occurs with many headers in ebcdic. To consider encoding
    errors, decode the textual header before passing it to wrap.

    Parameters
    ----------
    s : text or bytearray or str
    width : int

    Returns
    -------
    text : str

    Notes
    -----
    .. versionadded:: 1.1

    """
    try:
        s = s.decode(errors = 'ignore')
    except AttributeError:
        # Already str-like enough, e.g. wrap(f.text[0].decode()), so just try
        # to wrap-join
        pass

    return '\n'.join(textwrap.wrap(s, width=width))


def native(data,
           format = segyio.SegySampleFormat.IBM_FLOAT_4_BYTE,
           copy = True):
    """Convert numpy array to native float

    Converts a numpy array from raw segy trace data to native floats. Works for numpy ndarrays.

    Parameters
    ----------

    data : numpy.ndarray
    format : int or segyio.SegySampleFormat
    copy : bool
        If True, convert on a copy, and leave the input array unmodified

    Returns
    -------

    data : numpy.ndarray

    Notes
    -----

    .. versionadded:: 1.1

    Examples
    --------

    Convert mmap'd trace to native float:

    >>> d = np.memmap('file.sgy', offset = 3600, dtype = np.uintc)
    >>> samples = 1500
    >>> trace = segyio.tools.native(d[240:240+samples])

    """

    data = data.view( dtype = np.single )
    if copy:
        data = np.copy( data )

    format = int(segyio.SegySampleFormat(format))
    return segyio._segyio.native(data, format)

def collect(itr):
    """Collect traces or lines into one ndarray

    Eagerly copy a series of traces, lines or depths into one numpy ndarray. If
    collecting traces or fast-direction over a post-stacked file, reshaping the
    resulting array is equivalent to calling ``segyio.tools.cube``.

    Parameters
    ----------

    itr : iterable of numpy.ndarray

    Returns
    -------

    data : numpy.ndarray

    Notes
    -----

    .. versionadded:: 1.1

    Examples
    --------

    collect-cube identity:

    >>> with segyio.open('post-stack.sgy') as f:
    >>>     x = segyio.tools.collect(f.trace[:])
    >>>     x = x.reshape((len(f.ilines), len(f.xlines), f.samples))
    >>>     numpy.all(x == segyio.tools.cube(f))

    """
    return np.stack([np.copy(x) for x in itr])

def cube(f):
    """Read a full cube from a file

    Takes an open segy file (created with segyio.open) or a file name.

    If the file is a prestack file, the cube returned has the dimensions
    ``(fast, slow, offset, sample)``. If it is post-stack (only the one
    offset), the dimensions are normalised to ``(fast, slow, sample)``

    Parameters
    ----------

    f : str or segyio.SegyFile

    Returns
    -------

    cube : numpy.ndarray

    Notes
    -----

    .. versionadded:: 1.1

    """

    if not isinstance(f, segyio.SegyFile):
        with segyio.open(f) as fl:
            return cube(fl)

    ilsort = f.sorting == segyio.TraceSortingFormat.INLINE_SORTING
    fast = f.ilines if ilsort else f.xlines
    slow = f.xlines if ilsort else f.ilines
    fast, slow, offs = len(fast), len(slow), len(f.offsets)
    smps = len(f.samples)
    dims = (fast, slow, smps) if offs == 1 else (fast, slow, offs, smps)
    return f.trace.raw[:].reshape(dims)

def rotation(f, line = 'fast'):
    """ Find rotation of the survey

    Find the clock-wise rotation and origin of `line` as ``(rot, cdpx, cdpy)``

    The clock-wise rotation is defined as the angle in radians between line
    given by the first and last trace of the first line and the axis that gives
    increasing CDP-Y, in the direction that gives increasing CDP-X.

    By default, the first line is the 'fast' direction, which is inlines if the
    file is inline sorted, and crossline if it's crossline sorted.


    Parameters
    ----------

    f : SegyFile
    line : { 'fast', 'slow', 'iline', 'xline' }

    Returns
    -------

    rotation : float
    cdpx : int
    cdpy : int


    Notes
    -----

    .. versionadded:: 1.2

    """

    if f.unstructured:
        raise ValueError("Rotation requires a structured file")

    lines = { 'fast': f.fast,
              'slow': f.slow,
              'iline': f.iline,
              'xline': f.xline,
            }

    if line not in lines:
        error = "Unknown line {}".format(line)
        solution = "Must be any of: {}".format(' '.join(lines.keys()))
        raise ValueError('{} {}'.format(error, solution))

    l = lines[line]
    origin = f.header[0][segyio.su.cdpx, segyio.su.cdpy]
    cdpx, cdpy = origin[segyio.su.cdpx], origin[segyio.su.cdpy]

    rot = f.xfd.rotation( len(l),
                          l.stride,
                          len(f.offsets),
                          np.fromiter(l.keys(), dtype = np.intc) )
    return rot, cdpx, cdpy

def metadata(f):
    """Get survey structural properties and metadata

    Create a description object that, when passed to ``segyio.create()``, would
    create a new file with the same structure, dimensions, and metadata as
    ``f``.

    Takes an open segy file (created with segyio.open) or a file name.

    Parameters
    ----------

    f : str or segyio.SegyFile

    Returns
    -------
    spec : segyio.spec

    Notes
    -----

    .. versionadded:: 1.4

    """

    if not isinstance(f, segyio.SegyFile):
        with segyio.open(f) as fl:
            return metadata(fl)

    spec = segyio.spec()

    spec.iline = f._il
    spec.xline = f._xl
    spec.samples = f.samples
    spec.format = f.format

    spec.ilines = f.ilines
    spec.xlines = f.xlines
    spec.offsets = f.offsets
    spec.sorting = f.sorting

    spec.tracecount = f.tracecount

    spec.ext_headers = f.ext_headers
    spec.endian = f.endian

    return spec

def resample(f, rate = None, delay = None, micro = False,
                                           trace = True,
                                           binary = True):
    """Resample a file

    Resample all data traces, and update the file handle to reflect the new
    sample rate. No actual samples (data traces) are modified, only the header
    fields and interpretation.

    By default, the rate and the delay are in millseconds - if you need higher
    resolution, passing micro=True interprets rate as microseconds (as it is
    represented in the file). Delay is always milliseconds.

    By default, both the global binary header and the trace headers are updated
    to reflect this. If preserving either the trace header interval field or
    the binary header interval field is important, pass trace=False and
    binary=False respectively, to not have that field updated. This only apply
    to sample rates - the recording delay is only found in trace headers and
    will be written unconditionally, if delay is not None.

    .. warning::
        This function requires an open file handle and is **DESTRUCTIVE**. It
        will modify the file, and if an exception is raised then partial writes
        might have happened and the file might be corrupted.

    This function assumes all traces have uniform delays and frequencies.

    Parameters
    ----------

    f : SegyFile
    rate : int
    delay : int
    micro : bool
        if True, interpret rate as microseconds
    trace : bool
        Update the trace header if True
    binary : bool
        Update the binary header if True

    Notes
    -----

    .. versionadded:: 1.4

    """

    if rate is not None:
        if not micro: rate *= 1000

        if binary: f.bin[segyio.su.hdt] = rate
        if trace: f.header = { segyio.su.dt: rate}

    if delay is not None:
        f.header = { segyio.su.delrt: delay }

    t0 = delay if delay is not None else f.samples[0]
    rate = rate / 1000 if rate is not None else f.samples[1] - f.samples[0]

    f._samples = (np.arange(len(f.samples)) * rate) + t0

    return f


def from_array(filename, data, iline=189,
                               xline=193,
                               format=SegySampleFormat.IBM_FLOAT_4_BYTE,
                               dt=4000,
                               delrt=0):
    """ Create a new SEGY file from an n-dimentional array. Create a structured
    SEGY file with defaulted headers from a 2-, 3- or 4-dimensional array.
    ilines, xlines, offsets and samples are inferred from the size of the
    array. Please refer to the documentation for functions from_array2D,
    from_array3D and from_array4D to see how the arrays are interpreted.

    Structure-defining fields in the binary header and in the traceheaders are
    set accordingly. Such fields include, but are not limited to iline, xline
    and offset. The file also contains a defaulted textual header.

    Parameters
    ----------
    filename : string-like
        Path to new file
    data : 2-,3- or 4-dimensional array-like
    iline : int or segyio.TraceField
        Inline number field in the trace headers. Defaults to 189 as per the
        SEG-Y rev1 specification
    xline : int or segyio.TraceField
        Crossline number field in the trace headers. Defaults to 193 as per the
        SEG-Y rev1 specification
    format : int or segyio.SegySampleFormat
        Sample format field in the trace header. Defaults to IBM float 4 byte
    dt : int-like
        sample interval
    delrt : int-like

    Notes
    -----
    .. versionadded:: 1.8

    Examples
    --------
    Create a file from a 3D array, open it and read an iline:

    >>> segyio.tools.from_array(path, array3d)
    >>> segyio.open(path, mode) as f:
    ...     iline = f.iline[0]
    ...
    """

    dt = int(dt)
    delrt = int(delrt)

    data = np.asarray(data)
    dimensions = len(data.shape)

    if dimensions not in range(2, 5):
        problem = "Expected 2, 3, or 4 dimensions, {} was given".format(dimensions)
        raise ValueError(problem)

    spec = segyio.spec()
    spec.iline   = iline
    spec.xline   = xline
    spec.format  = format
    spec.sorting = TraceSortingFormat.INLINE_SORTING

    if dimensions == 2:
        spec.ilines  = [1]
        spec.xlines  = list(range(1, np.size(data,0) + 1))
        spec.samples = list(range(np.size(data,1)))
        spec.tracecount = np.size(data, 1)

    if dimensions == 3:
        spec.ilines  = list(range(1, np.size(data, 0) + 1))
        spec.xlines  = list(range(1, np.size(data, 1) + 1))
        spec.samples = list(range(np.size(data, 2)))

    if dimensions == 4:
        spec.ilines  = list(range(1, np.size(data, 0) + 1))
        spec.xlines  = list(range(1, np.size(data, 1) + 1))
        spec.offsets = list(range(1, np.size(data, 2)+ 1))
        spec.samples = list(range(np.size(data,3)))

    samplecount = len(spec.samples)

    with segyio.create(filename, spec) as f:
        tr = 0
        for ilno, il in enumerate(spec.ilines):
            for xlno, xl in enumerate(spec.xlines):
                for offno, off in enumerate(spec.offsets):
                    f.header[tr] = {
                        segyio.su.tracf  : tr,
                        segyio.su.cdpt   : tr,
                        segyio.su.offset : off,
                        segyio.su.ns     : samplecount,
                        segyio.su.dt     : dt,
                        segyio.su.delrt  : delrt,
                        segyio.su.iline  : il,
                        segyio.su.xline  : xl
                    }
                    if dimensions == 2: f.trace[tr] = data[tr, :]
                    if dimensions == 3: f.trace[tr] = data[ilno, xlno, :]
                    if dimensions == 4: f.trace[tr] = data[ilno, xlno, offno, :]
                    tr += 1

        f.bin.update(
            tsort=TraceSortingFormat.INLINE_SORTING,
            hdt=dt,
            dto=dt
        )


def from_array2D(filename, data, iline=189,
                                 xline=193,
                                 format=SegySampleFormat.IBM_FLOAT_4_BYTE,
                                 dt=4000,
                                 delrt=0):
    """ Create a new SEGY file from a 2D array
    Create an structured SEGY file with defaulted headers from a 2-dimensional
    array. The file is inline-sorted and structured as a slice, i.e. it has one
    iline and the xlinecount equals the tracecount. The tracecount and
    samplecount are inferred from the size of the array. Structure-defining
    fields in the binary header and in the traceheaders are set accordingly.
    Such fields include, but are not limited to iline, xline and offset. The
    file also contains a defaulted textual header.

    The 2 dimensional array is interpreted as::

                       samples
                 --------------------
        trace 0 | s0 | s1 | ... | sn |
                 --------------------
        trace 1 | s0 | s1 | ... | sn |
                 --------------------
                          .
                          .
                 --------------------
        trace n | s0 | s1 | ... | sn |
                 --------------------

    traces =  [0, len(axis(0)]
    samples = [0, len(axis(1)]

    Parameters
    ----------
    filename : string-like
        Path to new file
    data : 2-dimensional array-like
    iline : int or segyio.TraceField
        Inline number field in the trace headers. Defaults to 189 as per the
        SEG-Y rev1 specification
    xline : int or segyio.TraceField
        Crossline number field in the trace headers. Defaults to 193 as per the
        SEG-Y rev1 specification
    format : int or segyio.SegySampleFormat
        Sample format field in the trace header. Defaults to IBM float 4 byte
    dt : int-like
        sample interval
    delrt : int-like

    Notes
    -----
    .. versionadded:: 1.8

    Examples
    --------
    Create a file from a 2D array, open it and read a trace:

    >>> segyio.tools.from_array2D(path, array2d)
    >>> segyio.open(path, mode, strict=False) as f:
    ...     tr = f.trace[0]
    """

    data = np.asarray(data)
    dimensions = len(data.shape)

    if dimensions != 2:
        problem = "Expected 2 dimensions, {} was given".format(dimensions)
        raise ValueError(problem)

    from_array(filename, data, iline=iline, xline=xline, format=format,
                                                         dt=dt,
                                                         delrt=delrt)


def from_array3D(filename, data, iline=189,
                                 xline=193,
                                 format=SegySampleFormat.IBM_FLOAT_4_BYTE,
                                 dt=4000,
                                 delrt=0):
    """ Create a new SEGY file from a 3D array
    Create an structured SEGY file with defaulted headers from a 3-dimensional
    array. The file is inline-sorted. ilines, xlines and samples are inferred
    from the array. Structure-defining fields in the binary header and
    in the traceheaders are set accordingly. Such fields include, but are not
    limited to iline, xline and offset. The file also contains a defaulted
    textual header.

    The 3-dimensional array is interpreted as::

             xl0   xl1   xl2
            -----------------
         / | tr0 | tr1 | tr2 | il0
           -----------------
       | / | tr3 | tr4 | tr5 | il1
            -----------------
       | / | tr6 | tr7 | tr8 | il2
            -----------------
       | /      /     /     / n-samples
         ------------------

    ilines =  [1, len(axis(0) + 1]
    xlines =  [1, len(axis(1) + 1]
    samples = [0, len(axis(2)]

    Parameters
    ----------
    filename : string-like
        Path to new file
    data : 3-dimensional array-like
    iline : int or segyio.TraceField
        Inline number field in the trace headers. Defaults to 189 as per the
        SEG-Y rev1 specification
    xline : int or segyio.TraceField
        Crossline number field in the trace headers. Defaults to 193 as per the
        SEG-Y rev1 specification
    format : int or segyio.SegySampleFormat
        Sample format field in the trace header. Defaults to IBM float 4 byte
    dt : int-like
        sample interval
    delrt : int-like

    Notes
    -----
    .. versionadded:: 1.8

    Examples
    --------
    Create a file from a 3D array, open it and read an iline:

    >>> segyio.tools.from_array3D(path, array3d)
    >>> segyio.open(path, mode) as f:
    ...     iline = f.iline[0]
    ...
    """

    data = np.asarray(data)
    dimensions = len(data.shape)

    if dimensions != 3:
        problem = "Expected 3 dimensions, {} was given".format(dimensions)
        raise ValueError(problem)

    from_array(filename, data, iline=iline, xline=xline, format=format,
                                                         dt=dt,
                                                         delrt=delrt)


def from_array4D(filename, data, iline=189,
                                  xline=193,
                                  format=SegySampleFormat.IBM_FLOAT_4_BYTE,
                                  dt=4000,
                                  delrt=0):
    """ Create a new SEGY file from a 4D array
    Create an structured SEGY file with defaulted headers from a 4-dimensional
    array. The file is inline-sorted. ilines, xlines, offsets and samples are
    inferred from the array. Structure-defining fields in the binary header and
    in the traceheaders are set accordingly. Such fields include, but are not
    limited to iline, xline and offset. The file also contains a defaulted
    textual header.

    The 4D array is interpreted:

    ilines =  [1, len(axis(0) + 1]
    xlines =  [1, len(axis(1) + 1]
    offsets = [1, len(axis(2) + 1]
    samples = [0, len(axis(3)]

    Parameters
    ----------
    filename : string-like
        Path to new file
    data : 4-dimensional array-like
    iline : int or segyio.TraceField
        Inline number field in the trace headers. Defaults to 189 as per the
        SEG-Y rev1 specification
    xline : int or segyio.TraceField
        Crossline number field in the trace headers. Defaults to 193 as per the
        SEG-Y rev1 specification
    format : int or segyio.SegySampleFormat
        Sample format field in the trace header. Defaults to IBM float 4 byte
    dt : int-like
        sample interval
    delrt : int-like

    Notes
    -----
    .. versionadded:: 1.8

    Examples
    --------
    Create a file from a 3D array, open it and read an iline:

    >>> segyio.tools.create_from_array4D(path, array4d)
    >>> segyio.open(path, mode) as f:
    ...     iline = f.iline[0]
    ...
    """

    data = np.asarray(data)
    dimensions = len(data.shape)

    if dimensions != 4:
        problem = "Expected 4 dimensions, {} was given".format(dimensions)
        raise ValueError(problem)

    from_array(filename, data, iline=iline, xline=xline, format=format,
                                                         dt=dt,
                                                         delrt=delrt)



import collections
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping # noqa

import itertools
import numpy as np

try: from future_builtins import zip
except ImportError: pass

import segyio.tools as tools

class Gather(object):
    """
    A gather is in this context the intersection of lines in a cube, i.e. all
    the offsets at some inline/crossline intersection. The primary data type is
    numpy.ndarray, with dimensions depending on the range of offsets specified.

    Implements a dict_like lookup with the line and offset numbers (labels),
    not 0-based indices.

    Notes
    -----
    .. versionadded:: 1.1
    """

    def __init__(self, trace, iline, xline, offsets):
        # cache constructed modes for performance
        self.trace     = trace
        self.iline     = iline
        self.xline     = xline
        self.offsets   = offsets

    def __getitem__(self, index):
        """gather[i, x, o], gather[:,:,:]

        Get the gather or range of gathers, defined as offsets intersection
        between an in- and a crossline. Also works on post-stack files (with
        only 1 offset), although it is less useful in those cases.

        If offsets are omitted, the default is all offsets.

        A group of offsets is always returned as an offset-by-samples
        numpy.ndarray. If either inline, crossline, or both, are slices, a
        generator of such ndarrays are returned.

        If the slice of offsets misses all offsets, a special, empty ndarray is
        returned.

        Parameters
        ----------
        i : int or slice
            inline
        x : int or slice
            crossline
        o : int or slice
            offsets (default is :)

        Returns
        -------
        gather : numpy.ndarray or generator of numpy.ndarray

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------
        Read one offset at an intersection:

        >>> gather[200, 241, 25] # returns same shape as trace

        Read all offsets at an intersection:

        >>> gather[200, 241, :] # returns offsets x samples ndarray
        >>> # If no offset is specified, this is implicitly (:)
        >>> gather[200, 241, :] == gather[200, 241]

        All offsets for a set of ilines, intersecting one crossline:

        >>> gather[200:300, 241, :] == gather[200:300, 241]

        Some offsets for a set of ilines, interescting one crossline:

        >>> gather[200:300, 241, 10:25:5]

        Some offsets for a set of ilines and xlines. This effectively yields a
        subcube:

        >>> f.gather[200:300, 241:248, 1:10]
        """
        if len(index) < 3:
            index = (index[0], index[1], None)

        il, xl, off = index

        if off is None and len(self.offsets) == 1:
            off = self.offsets[0]

        # if offset isn't specified, default to all, [:]
        off = off or slice(None)

        def isslice(x): return isinstance(x, slice)

        # gather[int,int,int]
        if not any(map(isslice, [il, xl, off])):
            o = self.iline.offsets[off]
            i = o + self.iline.heads[il] + self.xline.heads[xl]
            return self.trace[i]

        offs = off if isslice(off) else slice(off, off+1, 1)

        xs = list(filter(self.offsets.__contains__,
                    range(*offs.indices(self.offsets[-1]+1))))

        empty = np.empty(0, dtype = self.trace.dtype)
        # gather[int,int,:]
        if not any(map(isslice, [il, xl])):
            if len(xs) == 0: return empty
            i = self.iline.heads[il] + self.xline.heads[xl]
            return tools.collect(self.trace[i + self.iline.offsets[x]] for x in xs)

        # gather[:,:,:], gather[int,:,:], gather[:,int,:]
        # gather[:,:,int] etc
        def gen():
            # precomputed xline number -> xline offset into the iline
            xlinds = { xlno: i for i, xlno in enumerate(self.xline.keys()) }

            # ranges over gathers are done on a by-line basis so lines can be
            # buffered, and traces can be read from the iline. This is the
            # least efficient when there are very few traces read per inline,
            # but huge savings with larger subcubes
            last_il = self.iline.keys()[-1] + 1
            last_xl = self.xline.keys()[-1] + 1

            il_slice = il if isslice(il) else slice(il, il+1)
            xl_slice = xl if isslice(xl) else slice(xl, xl+1)

            # the slice could still be defaulted (:), in which case this starts
            # at zero. the lookups are guarded by try-excepts so it won't fail,
            # but it's unnecessary to chck all keys up until the first xline
            # because that will never be a hit anyway
            if il_slice.start is None:
                start = self.iline.keys()[0]
                il_slice = slice(start, il_slice.stop, il_slice.step)

            if xl_slice.start is None:
                start = self.xline.keys()[0]
                xl_slice = slice(start, xl_slice.stop, xl_slice.step)

            il_range = range(*il_slice.indices(last_il))
            xl_range = range(*xl_slice.indices(last_xl))

            # the try-except-else is essentially a filter on in/xl keys, but
            # delegates the work (and decision) to the iline and xline modes
            if not isslice(off):
                for iline in self.iline[il_slice, off]:
                    for xlno in xl_range:
                        try:
                            xind = xlinds[xlno]
                        except KeyError:
                            pass
                        else:
                            yield iline[xind]

                return

            if len(xs) == 0:
                for _, _ in itertools.product(il_range, xl_range): yield empty
                return

            for ilno in il_range:
                iline = tools.collect(self.iline[ilno, off])
                for x in xl_range:
                    try:
                        xind = xlinds[x]
                    except KeyError:
                        pass
                    else:
                        yield iline[:, xind]

        return gen()

class Group(object):
    """
    The inner representation of the Groups abstraction provided by Group.

    A collection of trace indices that have identical `key`.

    Notes
    -----
    .. versionadded:: 1.9
    """
    def __init__(self, key, parent, index):
        self.parent = parent
        self.index = index
        self.key = key

    @property
    def header(self):
        """
        A generator of the the read-only headers in this group

        Returns
        -------
        headers : iterator of Header

        Notes
        -----
        The generator respects the order of the index - to iterate over headers
        in a different order, the index attribute can be re-organised.

        .. versionadded:: 1.9
        """
        source = self.parent.header
        for i in self.index:
            yield source[i]

    @property
    def trace(self):
        """
        A generator of the the read-only traces in this group

        Returns
        -------
        traces : iterator of Trace

        Notes
        -----
        The generator respects the order of the index - to iterate over headers
        in a different order, the index attribute can be re-organised.

        .. versionadded:: 1.9
        """
        source = self.parent.trace
        for i in self.index:
            yield source[i]

    def sort(self, fields):
        """
        Sort the traces in the group, obeying the `fields` order of
        most-to-least significant word.
        """
        # TODO: examples

        headers = [dict(self.parent.header[i]) for i in self.index]
        index = list(zip(headers, self.index))
        # sorting is stable, so sort the whole set by field, applied in the
        # reverse order:
        for field in reversed(fields):
            index.sort(key = lambda x: x[0][field])

        # strip off all the headers
        index = [i for _, i in index]
        self.index = index

class Groups(Mapping):
    """
    The Groups implements the dict interface, grouping all traces that match a
    given `fingerprint`. The fingerprint is a signature derived from a set of
    trace header words, called a `key`.

    Consider a file with five traces, and some selected header words::

        0: {offset: 1, fldr: 1}
        1: {offset: 1, fldr: 2}
        2: {offset: 1, fldr: 1}
        3: {offset: 2, fldr: 1}
        4: {offset: 1, fldr: 2}

    With key = (offset, fldr), there are 3 groups::

        {offset: 1, fldr: 1 }: [0, 2]
        {offset: 1, fldr: 2 }: [1, 4]
        {offset: 2, fldr: 1 }: [3]

    With a key = offset, there are 2 groups::

        {offset: 1}: [0, 1, 2, 4]
        {offset: 2}: [3]

    The Groups class is intended to easily process files without the rigid
    in/crossline structure of iline/xline/gather, but where there is sufficient
    structure in the headers. This is common for some types of pre-stack data,
    shot gather data etc.

    Notes
    -----
    .. versionadded:: 1.9
    """
    # TODO: only group in range of traces?
    # TODO: cache header dicts?
    def __init__(self, trace, header, key):
        bins = collections.OrderedDict()
        for i, h in enumerate(header[:]):
            k = self.fingerprint(h[key])
            if k in bins:
                bins[k].append(i)
            else:
                bins[k] = [i]

        self.trace = trace
        self.header = header
        self.key = key
        self.bins = bins

    @staticmethod
    def normalize_keys(items):
        """
        Normalize the key representation to integers, so that they're hashable,
        even when a key is built with enumerators.

        This function is intended for internal use, and provides the mapping
        from accepted key representation to a canonical key.

        Parameters
        ----------
        items : iterator of (int_like, array_like)

        Returns
        -------
        items : generator of (int, array_like)

        Warnings
        --------
        This function provides no guarantees for value and type compatibility,
        even between minor versions.

        Notes
        -----
        .. versionadded:: 1.9
        """
        return ((int(k), v) for k, v in items)

    @staticmethod
    def fingerprint(key):
        """
        Compute a hashable fingerprint for a key. This function is intended for
        internal use. Relies on normalize_keys for transforming keys to
        canonical form. The output of this function is used for the group ->
        index mapping.

        Parameters
        ----------
        key : int_like or dict of {int_like: int} or iterable of (int_like,int)

        Returns
        -------
        key
            A normalized canonical representation of key

        Warnings
        --------
        This function provides no guarantees for value and type compatibility,
        even between minor versions.

        Notes
        -----
        .. versionadded:: 1.9
        """
        try:
            return int(key)
        except TypeError:
            pass

        try:
            items = key.items()
        except AttributeError:
            items = iter(key)

        # map k -> tracefield -> int
        return frozenset(Groups.normalize_keys(items))

    def __len__(self):
        """x.__len__() <==> len(x)"""
        return len(self.bins)

    def __contains__(self, key):
        """x.__len__() <==> len(x)"""
        return self.fingerprint(key) in self.bins

    def __getitem__(self, key):
        """g[key]

        Read the group associated with key.

        Key can be any informal mapping between a header word (TraceField, su
        header words, or raw integers) and a value.

        Parameters
        ----------
        key

        Returns
        -------
        group : Group

        Notes
        -----
        .. versionadded:: 1.9

        Examples
        --------

        Group on FieldRecord, and get the group FieldRecord == 5:

        >>> fr = segyio.TraceField.FieldRecord
        >>> records = f.group(fr)
        >>> record5 = records[5]
        """
        key = self.fingerprint(key)
        return Group(key, self, self.bins[key])

    def values(self):
        for key, index in self.bins.items():
            yield Group(key, self, index)

    def items(self):
        for key, index in self.bins.items():
            yield key, Group(key, self, index)

    def __iter__(self):
        return self.bins.keys()

    def sort(self, fields):
        """
        Reorganise the indices in all groups by fields
        """
        bins = collections.OrderedDict()

        for key, index in self.bins.items():
            g = Group(key, self, index)
            g.sort(fields)
            bins[key] = g.index

        self.bins = bins



try:
    from collections.abc import Mapping # noqa
    from collections.abc import MutableMapping # noqa
except ImportError:
    from collections import Mapping # noqa
    from collections import MutableMapping # noqa

import segyio
from .binfield import BinField
from .tracefield import TraceField

class Field(MutableMapping):
    """
    The Field implements the dict interface, with a fixed set of keys. It's
    used for both binary- and trace headers. Any modifications to this
    dict_like object will be reflected on disk.

    The keys can be integers, int_likes, or enumerations such as BinField,
    TraceField, and su. If raw, numerical offsets are used they must align with
    the defined byte offsets by the SEGY specification.

    Notes
    -----
    .. versionadded:: 1.1

    .. versionchanged:: 1.3
        common dict operations (update, keys, values)

    .. versionchanged:: 1.6
        more common dict operations (MutableMapping)
    """
    _bin_keys = [x for x in BinField.enums()
                 if  x != BinField.Unassigned1
                 and x != BinField.Unassigned2]

    _tr_keys = [x for x in TraceField.enums()
                if  x != TraceField.UnassignedInt1
                and x != TraceField.UnassignedInt2]

    _kwargs = {
        'tracl' : TraceField.TRACE_SEQUENCE_LINE,
        'tracr' : TraceField.TRACE_SEQUENCE_FILE,
        'fldr'  : TraceField.FieldRecord,
        'tracf' : TraceField.TraceNumber,
        'ep'    : TraceField.EnergySourcePoint,
        'cdp'   : TraceField.CDP,
        'cdpt'  : TraceField.CDP_TRACE,
        'trid'  : TraceField.TraceIdentificationCode,
        'nvs'   : TraceField.NSummedTraces,
        'nhs'   : TraceField.NStackedTraces,
        'duse'  : TraceField.DataUse,
        'offset': TraceField.offset,
        'gelev' : TraceField.ReceiverGroupElevation,
        'selev' : TraceField.SourceSurfaceElevation,
        'sdepth': TraceField.SourceDepth,
        'gdel'  : TraceField.ReceiverDatumElevation,
        'sdel'  : TraceField.SourceDatumElevation,
        'swdep' : TraceField.SourceWaterDepth,
        'gwdep' : TraceField.GroupWaterDepth,
        'scalel': TraceField.ElevationScalar,
        'scalco': TraceField.SourceGroupScalar,
        'sx'    : TraceField.SourceX,
        'sy'    : TraceField.SourceY,
        'gx'    : TraceField.GroupX,
        'gy'    : TraceField.GroupY,
        'counit': TraceField.CoordinateUnits,
        'wevel' : TraceField.WeatheringVelocity,
        'swevel': TraceField.SubWeatheringVelocity,
        'sut'   : TraceField.SourceUpholeTime,
        'gut'   : TraceField.GroupUpholeTime,
        'sstat' : TraceField.SourceStaticCorrection,
        'gstat' : TraceField.GroupStaticCorrection,
        'tstat' : TraceField.TotalStaticApplied,
        'laga'  : TraceField.LagTimeA,
        'lagb'  : TraceField.LagTimeB,
        'delrt' : TraceField.DelayRecordingTime,
        'muts'  : TraceField.MuteTimeStart,
        'mute'  : TraceField.MuteTimeEND,
        'ns'    : TraceField.TRACE_SAMPLE_COUNT,
        'dt'    : TraceField.TRACE_SAMPLE_INTERVAL,
        'gain'  : TraceField.GainType,
        'igc'   : TraceField.InstrumentGainConstant,
        'igi'   : TraceField.InstrumentInitialGain,
        'corr'  : TraceField.Correlated,
        'sfs'   : TraceField.SweepFrequencyStart,
        'sfe'   : TraceField.SweepFrequencyEnd,
        'slen'  : TraceField.SweepLength,
        'styp'  : TraceField.SweepType,
        'stat'  : TraceField.SweepTraceTaperLengthStart,
        'stae'  : TraceField.SweepTraceTaperLengthEnd,
        'tatyp' : TraceField.TaperType,
        'afilf' : TraceField.AliasFilterFrequency,
        'afils' : TraceField.AliasFilterSlope,
        'nofilf': TraceField.NotchFilterFrequency,
        'nofils': TraceField.NotchFilterSlope,
        'lcf'   : TraceField.LowCutFrequency,
        'hcf'   : TraceField.HighCutFrequency,
        'lcs'   : TraceField.LowCutSlope,
        'hcs'   : TraceField.HighCutSlope,
        'year'  : TraceField.YearDataRecorded,
        'day'   : TraceField.DayOfYear,
        'hour'  : TraceField.HourOfDay,
        'minute': TraceField.MinuteOfHour,
        'sec'   : TraceField.SecondOfMinute,
        'timbas': TraceField.TimeBaseCode,
        'trwf'  : TraceField.TraceWeightingFactor,
        'grnors': TraceField.GeophoneGroupNumberRoll1,
        'grnofr': TraceField.GeophoneGroupNumberFirstTraceOrigField,
        'grnlof': TraceField.GeophoneGroupNumberLastTraceOrigField,
        'gaps'  : TraceField.GapSize,
        'otrav' : TraceField.OverTravel,
        'cdpx'  : TraceField.CDP_X,
        'cdpy'  : TraceField.CDP_Y,
        'iline' : TraceField.INLINE_3D,
        'xline' : TraceField.CROSSLINE_3D,
        'sp'    : TraceField.ShotPoint,
        'scalsp': TraceField.ShotPointScalar,
        'trunit': TraceField.TraceValueMeasurementUnit,
        'tdcm'  : TraceField.TransductionConstantMantissa,
        'tdcp'  : TraceField.TransductionConstantPower,
        'tdunit': TraceField.TransductionUnit,
        'triden': TraceField.TraceIdentifier,
        'sctrh' : TraceField.ScalarTraceHeader,
        'stype' : TraceField.SourceType,
        'sedm'  : TraceField.SourceEnergyDirectionMantissa,
        'sede'  : TraceField.SourceEnergyDirectionExponent,
        'smm'   : TraceField.SourceMeasurementMantissa,
        'sme'   : TraceField.SourceMeasurementExponent,
        'smunit': TraceField.SourceMeasurementUnit,
        'uint1' : TraceField.UnassignedInt1,
        'uint2' : TraceField.UnassignedInt2,

        'jobid' : BinField.JobID,
        'lino'  : BinField.LineNumber,
        'reno'  : BinField.ReelNumber,
        'ntrpr' : BinField.Traces,
        'nart'  : BinField.AuxTraces,
        'hdt'   : BinField.Interval,
        'dto'   : BinField.IntervalOriginal,
        'hns'   : BinField.Samples,
        'nso'   : BinField.SamplesOriginal,
        'format': BinField.Format,
        'fold'  : BinField.EnsembleFold,
        'tsort' : BinField.SortingCode,
        'vscode': BinField.VerticalSum,
        'hsfs'  : BinField.SweepFrequencyStart,
        'hsfe'  : BinField.SweepFrequencyEnd,
        'hslen' : BinField.SweepLength,
        'hstyp' : BinField.Sweep,
        'schn'  : BinField.SweepChannel,
        'hstas' : BinField.SweepTaperStart,
        'hstae' : BinField.SweepTaperEnd,
        'htatyp': BinField.Taper,
        'hcorr' : BinField.CorrelatedTraces,
        'bgrcv' : BinField.BinaryGainRecovery,
        'rcvm'  : BinField.AmplitudeRecovery,
        'mfeet' : BinField.MeasurementSystem,
        'polyt' : BinField.ImpulseSignalPolarity,
        'vpol'  : BinField.VibratoryPolarity,
        'extntrpr' : BinField.ExtTraces,
        'extnart'  : BinField.ExtAuxTraces,
        'exthns'   : BinField.ExtSamples,
        'extnso'   : BinField.ExtSamplesOriginal,
        'extfold'  : BinField.ExtEnsembleFold,
        'unas1' : BinField.Unassigned1,
        'rev'   : BinField.SEGYRevision,
        'revmin': BinField.SEGYRevisionMinor,
        'trflag': BinField.TraceFlag,
        'exth'  : BinField.ExtendedHeaders,
        'unas2' : BinField.Unassigned2,
    }

    def __init__(self, buf, kind, traceno = None, filehandle = None, readonly = True):
        # do setup of kind/keys first, so that keys() work. if this method
        # throws, we want repr() to be well-defined for backtrace, and that
        # requires _keys
        if kind == 'binary':
            self._keys = self._bin_keys
            self.kind = BinField
        elif kind == 'trace':
            self._keys = self._tr_keys
            self.kind = TraceField
        else:
            raise ValueError('Unknown header type {}'.format(kind))

        self.buf = buf
        self.traceno = traceno
        self.filehandle = filehandle
        self.getfield = segyio._segyio.getfield
        self.putfield = segyio._segyio.putfield

        self.readonly = readonly

    def fetch(self, buf = None, traceno = None):
        """Fetch the header from disk

        This object will read header when it is constructed, which means it
        might be out-of-date if the file is updated through some other handle.
        This method is largely meant for internal use - if you need to reload
        disk contents, use ``reload``.

        Fetch does not update any internal state (unless `buf` is ``None`` on a
        trace header, and the read succeeds), but returns the fetched header
        contents.

        This method can be used to reposition the trace header, which is useful
        for constructing generators.

        If this is called on a writable, new file, and this header has not yet
        been written to, it will successfully return an empty buffer that, when
        written to, will be reflected on disk.

        Parameters
        ----------
        buf     : bytearray
            buffer to read into instead of ``self.buf``
        traceno : int

        Returns
        -------
        buf : bytearray

        Notes
        -----
        .. versionadded:: 1.6

        This method is not intended as user-oriented functionality, but might
        be useful in high-performance code.
        """

        if buf is None:
            buf = self.buf

        if traceno is None:
            traceno = self.traceno

        try:
            if self.kind == TraceField:
                if traceno is None: return buf
                return self.filehandle.getth(traceno, buf)
            else:
                return self.filehandle.getbin()
        except IOError:
            if not self.readonly:
                # the file was probably newly created and the trace header
                # hasn't been written yet, and we set the buffer to zero. if
                # this is the case we want to try and write it later, and if
                # the file was broken, permissions were wrong etc writing will
                # fail too
                #
                # if the file is opened read-only and this happens, there's no
                # way to actually write and the error is an actual error
                return bytearray(len(self.buf))
            else: raise

    def reload(self):
        """
        This object will read header when it is constructed, which means it
        might be out-of-date if the file is updated through some other handle.

        It's rarely required to call this method, and it's a symptom of fragile
        code. However, if you have multiple handles to the same header, it
        might be necessary. Consider the following example::

            >>> x = f.header[10]
            >>> y = f.header[10]
            >>> x[1, 5]
            { 1: 5, 5: 10 }
            >>> y[1, 5]
            { 1: 5, 5: 10 }
            >>> x[1] = 6
            >>> x[1], y[1] # write to x[1] is invisible to y
            6, 5
            >>> y.reload()
            >>> x[1], y[1]
            6, 6
            >>> x[1] = 5
            >>> x[1], y[1]
            5, 6
            >>> y[5] = 1
            >>> x.reload()
            >>> x[1], y[1, 5] # the write to x[1] is lost
            6, { 1: 6; 5: 1 }

        In segyio, headers writes are atomic, and the write to disk writes the
        full cache. If this cache is out of date, some writes might get lost,
        even though the updates are compatible.

        The fix to this issue is either to use ``reload`` and maintain buffer
        consistency, or simply don't let header handles alias and overlap in
        lifetime.

        Notes
        -----
        .. versionadded:: 1.6
        """

        self.buf = self.fetch(buf = self.buf)
        return self

    def flush(self):
        """Commit backing storage to disk

        This method is largely internal, and it is not necessary to call this
        from user code. It should not be explicitly invoked and may be removed
        in future versions.
        """

        if self.kind == TraceField:
            self.filehandle.putth(self.traceno, self.buf)

        elif self.kind == BinField:
            self.filehandle.putbin(self.buf)

        else:
            msg = 'Object corrupted: kind {} not valid'
            raise RuntimeError(msg.format(self.kind))

    def __getitem__(self, key):
        """d[key]

        Read the associated value of `key`.

        `key` can be any iterable, to retrieve multiple keys at once. In this
        case, a mapping of key -> value is returned.

        Parameters
        ----------
        key : int, or iterable of int

        Returns
        -------
        value : int or dict_like

        Notes
        -----
        .. versionadded:: 1.1

        .. note::
            Since version 1.6, KeyError is appropriately raised on key misses,
            whereas ``IndexError`` was raised before. This is an old bug, since
            header types were documented to be dict-like. If you rely on
            catching key-miss errors in your code, you might want to handle
            both ``IndexError`` and ``KeyError`` for multi-version robustness.

        .. warning::
            segyio considers reads/writes full headers, not individual fields,
            and does the read from disk when this class is constructed. If the
            file is updated through some other handle, including a secondary
            access via `f.header`, this cache might be out-of-date.

        Examples
        --------
        Read a single value:

        >>> d[3213]
        15000

        Read multiple values at once:

        >>> d[37, 189]
        { 37: 5, 189: 2484 }
        >>> d[37, TraceField.INLINE_3D]
        { 37: 5, 189: 2484 }
        """

        try: return self.getfield(self.buf, int(key))
        except TypeError: pass

        return {self.kind(k): self.getfield(self.buf, int(k)) for k in key}

    def __setitem__(self, key, val):
        """d[key] = val

        Set d[key] to val. Setting keys commits changes to disk, although the
        changes may not be visible until the kernel schedules the write.

        Unlike d[key], this method does not support assigning multiple values
        at once. To set multiple values at once, use the `update` method.

        Parameters
        ----------
        key : int_like
        val : int_like

        Returns
        -------
        val : int
            The value set

        Notes
        -----
        .. versionadded:: 1.1

        .. note::
            Since version 1.6, KeyError is appropriately raised on key misses,
            whereas ``IndexError`` was raised before. This is an old bug, since
            header types were documented to be dict-like. If you rely on
            catching key-miss errors in your code, you might want to handle
            both ``IndexError`` and ``KeyError`` for multi-version robustness.

        .. warning::
            segyio considers reads/writes full headers, not individual fields,
            and does the read from disk when this class is constructed. If the
            file is updated through some other handle, including a secondary
            access via `f.header`, this cache might be out-of-date. That means
            writing an individual field will write the full header to disk,
            possibly overwriting previously set values.

        Examples
        --------
        Set a value and keep in a variable:

        >>> x = header[189] = 5
        >>> x
        5
        """

        self.putfield(self.buf, key, val)
        self.flush()

        return val

    def __delitem__(self, key):
        """del d[key]

        'Delete' the key by setting value to zero. Equivalent to ``d[key] =
        0``.

        Notes
        -----
        .. versionadded:: 1.6
        """

        self[key] = 0

    def keys(self):
        """D.keys() -> a set-like object providing a view on D's keys"""
        return list(self._keys)

    def __len__(self):
        """x.__len__() <==> len(x)"""
        return len(self._keys)

    def __iter__(self):
        """x.__iter__() <==> iter(x)"""
        return iter(self._keys)

    def __eq__(self, other):
        """x.__eq__(y) <==> x == y"""

        if not isinstance(other, Mapping):
            return NotImplemented

        if len(self) != len(other):
            return False

        def intkeys(d):
            return { int(k): v for k, v in d.items() }

        return intkeys(self) == intkeys(other)


    def update(self, *args, **kwargs):
        """d.update([E, ]**F) -> None.  Update D from mapping/iterable E and F.

        Overwrite the values in `d` with the keys from `E` and `F`. If any key
        in `value` is invalid in `d`, ``KeyError`` is raised.

        This method is atomic - either all values in `value` are set in `d`, or
        none are. ``update`` does not commit a partially-updated version to
        disk.

        For kwargs, Seismic Unix-style names are supported. `BinField` and
        `TraceField` are not, because there are name collisions between them,
        although this restriction may be lifted in the future.

        Notes
        -----
        .. versionchanged:: 1.3
            Support for common dict operations (update, keys, values)

        .. versionchanged:: 1.6
            Atomicity guarantee

        .. versionchanged:: 1.6
            `**kwargs` support

        Examples
        --------
        >>> e = { 1: 10, 9: 5 }
        >>> d.update(e)
        >>> l = [ (105, 11), (169, 4) ]
        >>> d.update(l)
        >>> d.update(e, iline=189, xline=193, hour=5)
        >>> d.update(sx=7)

        """

        if len(args) > 1:
            msg = 'update expected at most 1 non-keyword argument, got {}'
            raise TypeError(msg.format(len(args)))

        buf = bytearray(self.buf)

        # Implementation largely borrowed from Mapping
        # If E present and has a .keys() method: for k in E: D[k] = E[k]
        # If E present and lacks .keys() method: for (k, v) in E: D[k] = v
        # In either case, this is followed by: for k, v in F.items(): D[k] = v
        if len(args) == 1:
            other = args[0]
            if isinstance(other, Mapping):
                for key in other:
                    self.putfield(buf, int(key), other[key])
            elif hasattr(other, "keys"):
                for key in other.keys():
                    self.putfield(buf, int(key), other[key])
            else:
                for key, value in other:
                    self.putfield(buf, int(key), value)

        for key, value in kwargs.items():
            self.putfield(buf, int(self._kwargs[key]), value)

        self.buf = buf
        self.flush()

    @classmethod
    def binary(cls, segy):
        buf = bytearray(segyio._segyio.binsize())
        return Field(buf, kind='binary',
                          filehandle=segy.xfd,
                          readonly=segy.readonly,
                    ).reload()

    @classmethod
    def trace(cls, traceno, segy):
        buf = bytearray(segyio._segyio.thsize())
        return Field(buf, kind='trace',
                          traceno=traceno,
                          filehandle=segy.xfd,
                          readonly=segy.readonly,
                    ).reload()

    def __repr__(self):
        return repr(self[self.keys()])



import datetime
import numpy
import segyio

from . import TraceSortingFormat

def default_text_header(iline, xline, offset):
    lines = {
        1: "DATE %s" % datetime.date.today().isoformat(),
        2: "AN INCREASE IN AMPLITUDE EQUALS AN INCREASE IN ACOUSTIC IMPEDANCE",
        3: "Written by libsegyio (python)",
        11: "TRACE HEADER POSITION:",
        12: "  INLINE BYTES %03d-%03d    | OFFSET BYTES %03d-%03d" % (iline, iline + 4, int(offset), int(offset) + 4),
        13: "  CROSSLINE BYTES %03d-%03d |" % (xline, xline + 4),
        15: "END EBCDIC HEADER",
    }
    rows = segyio.create_text_header(lines)
    rows = bytearray(rows, 'ascii')  # mutable array of bytes
    rows[-1] = 128  # \x80 -- Unsure if this is really required...
    return bytes(rows)  # immutable array of bytes that is compatible with strings


def structured(spec):
    if not hasattr(spec, 'ilines' ): return False
    if not hasattr(spec, 'xlines' ): return False
    if not hasattr(spec, 'offsets'): return False

    if spec.ilines  is None: return False
    if spec.xlines  is None: return False
    if spec.offsets is None: return False

    if not list(spec.ilines):  return False
    if not list(spec.xlines):  return False
    if not list(spec.offsets): return False

    return True

def create(filename, spec):
    """Create a new segy file.

    Create a new segy file with the geometry and properties given by `spec`.
    This enables creating SEGY files from your data. The created file supports
    all segyio modes, but has an emphasis on writing. The spec must be
    complete, otherwise an exception will be raised. A default, empty spec can
    be created with ``segyio.spec()``.

    Very little data is written to the file, so just calling create is not
    sufficient to re-read the file with segyio. Rather, every trace header and
    trace must be written to the file to be considered complete.

    Create should be used together with python's ``with`` statement. This ensure
    the data is written. Please refer to the examples.

    The ``segyio.spec()`` function will default sorting, offsets and everything
    in the mandatory group, except format and samples, and requires the caller
    to fill in *all* the fields in either of the exclusive groups.

    If any field is missing from the first exclusive group, and the tracecount
    is set, the resulting file will be considered unstructured. If the
    tracecount is set, and all fields of the first exclusive group are
    specified, the file is considered structured and the tracecount is inferred
    from the xlines/ilines/offsets. The offsets are defaulted to ``[1]`` by
    ``segyio.spec()``.

    Parameters
    ----------
    filename : str
        Path to file to create
    spec : segyio.spec
        Structure of the segy file

    Returns
    -------
    file : segyio.SegyFile
        An open segyio file handle, similar to that returned by `segyio.open`

    See also
    --------
    segyio.spec : template for the `spec` argument


    Notes
    -----

    .. versionadded:: 1.1

    .. versionchanged:: 1.4
       Support for creating unstructured files

    .. versionchanged:: 1.8
       Support for creating lsb files

    The ``spec`` is any object that has the following attributes

    Mandatory::

        iline   : int or segyio.BinField
        xline   : int or segyio.BinField
        samples : array of int
        format  : { 1, 5 }
            1 = IBM float, 5 = IEEE float

    Exclusive::

        ilines  : array_like of int
        xlines  : array_like of int
        offsets : array_like of int
        sorting : int or segyio.TraceSortingFormat

        OR

        tracecount : int

    Optional::

        ext_headers : int
        endian : str { 'big', 'msb', 'little', 'lsb' }
            defaults to 'big'


    Examples
    --------

    Create a file:

    >>> spec = segyio.spec()
    >>> spec.ilines  = [1, 2, 3, 4]
    >>> spec.xlines  = [11, 12, 13]
    >>> spec.samples = list(range(50))
    >>> spec.sorting = 2
    >>> spec.format  = 1
    >>> with segyio.create(path, spec) as f:
    ...     ## fill the file with data
    ...     pass
    ...

    Copy a file, but shorten all traces by 50 samples:

    >>> with segyio.open(srcpath) as src:
    ...     spec = segyio.spec()
    ...     spec.sorting = src.sorting
    ...     spec.format = src.format
    ...     spec.samples = src.samples[:len(src.samples) - 50]
    ...     spec.ilines = src.ilines
    ...     spec.xline = src.xlines
    ...     with segyio.create(dstpath, spec) as dst:
    ...         dst.text[0] = src.text[0]
    ...         dst.bin = src.bin
    ...         # this is writing a sparse file, which might be slow on some
    ...         # systems
    ...         dst.header = src.header
    ...         dst.trace = src.trace

     Copy a file, but shift samples time by 50:

    >>> with segyio.open(srcpath) as src:
    ...     delrt = 50
    ...     spec = segyio.spec()
    ...     spec.samples = src.samples + delrt
    ...     spec.ilines = src.ilines
    ...     spec.xline = src.xlines
    ...     with segyio.create(dstpath, spec) as dst:
    ...         dst.text[0] = src.text[0]
    ...         dst.bin = src.bin
    ...         dst.header = src.header
    ...         dst.header = { TraceField.DelayRecordingTime: delrt }
    ...         dst.trace = src.trace

    Copy a file, but shorten all traces by 50 samples (since v1.4):

    >>> with segyio.open(srcpath) as src:
    ...     spec = segyio.tools.metadata(src)
    ...     spec.samples = spec.samples[:len(spec.samples) - 50]
    ...     with segyio.create(dstpath, spec) as dst:
    ...         dst.text[0] = src.text[0]
    ...         dst.bin = src.bin
    ...         dst.header = src.header
    ...         dst.trace = src.trace
    """

    from . import _segyio

    if not structured(spec):
        tracecount = spec.tracecount
    else:
        tracecount = len(spec.ilines) * len(spec.xlines) * len(spec.offsets)

    ext_headers = spec.ext_headers if hasattr(spec, 'ext_headers') else 0
    samples = numpy.asarray(spec.samples)

    endians = {
        'lsb': 256, # (1 << 8)
        'little': 256,
        'msb': 0,
        'big': 0,
    }
    endian = spec.endian if hasattr(spec, 'endian') else 'big'
    if endian is None:
        endian = 'big'

    if endian not in endians:
        problem = 'unknown endianness {}, expected one of: '
        opts = ' '.join(endians.keys())
        raise ValueError(problem.format(endian) + opts)

    fd = _segyio.segyiofd(str(filename), 'w+', endians[endian])
    fd.segymake(
        samples = len(samples),
        tracecount = tracecount,
        format = int(spec.format),
        ext_headers = int(ext_headers),
    )

    f = segyio.SegyFile(fd,
            filename = str(filename),
            mode = 'w+',
            iline = int(spec.iline),
            xline = int(spec.xline),
            endian = endian,
    )

    f._samples       = samples

    if structured(spec):
        sorting = spec.sorting if hasattr(spec, 'sorting') else None
        if sorting is None:
            sorting = TraceSortingFormat.INLINE_SORTING
        f.interpret(spec.ilines, spec.xlines, spec.offsets, sorting)

    f.text[0] = default_text_header(f._il, f._xl, segyio.TraceField.offset)

    if len(samples) == 1:
        interval = int(samples[0] * 1000)
    else:
        interval = int((samples[1] - samples[0]) * 1000)

    f.bin.update(
        ntrpr    = tracecount,
        nart     = tracecount,
        hdt      = interval,
        dto      = interval,
        hns      = len(samples),
        nso      = len(samples),
        format   = int(spec.format),
        exth     = ext_headers,
    )

    if len(samples) > 2**16 - 1:
        # when using the ext-samples field, also set rev2, even though it's a
        # soft lie and files aren't really compliant
        f.bin.update(
            exthns = len(samples),
            extnso = len(samples),
            rev    = 2
        )

    return f



import numpy

import segyio

def infer_geometry(f, metrics, iline, xline, strict):
    try:
        cube_metrics = f.xfd.cube_metrics(iline, xline)
        f._sorting   = cube_metrics['sorting']
        iline_count  = cube_metrics['iline_count']
        xline_count  = cube_metrics['xline_count']
        offset_count = cube_metrics['offset_count']
        metrics.update(cube_metrics)

        ilines  = numpy.zeros(iline_count,  dtype=numpy.intc)
        xlines  = numpy.zeros(xline_count,  dtype=numpy.intc)
        offsets = numpy.zeros(offset_count, dtype=numpy.intc)

        f.xfd.indices(metrics, ilines, xlines, offsets)
        f.interpret(ilines, xlines, offsets, f._sorting)

    except:
        if not strict:
            f._ilines  = None
            f._xlines  = None
            f._offsets = None
        else:
            f.close()
            raise

    return f


def open(filename, mode="r", iline = 189,
                             xline = 193,
                             strict = True,
                             ignore_geometry = False,
                             endian = 'big'):
    """Open a segy file.

    Opens a segy file and tries to figure out its sorting, inline numbers,
    crossline numbers, and offsets, and enables reading and writing to this
    file in a simple manner.

    For reading, the access mode `r` is preferred. All write operations will
    raise an exception. For writing, the mode `r+` is preferred (as `rw` would
    truncate the file). Any mode with `w` will raise an error. The modes used
    are standard C file modes; please refer to that documentation for a
    complete reference.

    Open should be used together with python's ``with`` statement. Please refer
    to the examples. When the ``with`` statement is used the file will
    automatically be closed when the routine completes or an exception is
    raised.

    By default, segyio tries to open in ``strict`` mode. This means the file will
    be assumed to represent a geometry with consistent inline, crosslines and
    offsets. If strict is False, segyio will still try to establish a geometry,
    but it won't abort if it fails. When in non-strict mode is opened,
    geometry-dependent modes such as iline will raise an error.

    If ``ignore_geometry=True``, segyio will *not* try to build iline/xline or
    other geometry related structures, which leads to faster opens. This is
    essentially the same as using ``strict=False`` on a file that has no
    geometry.

    Parameters
    ----------

    filename : str
        Path to file to open

    mode : {'r', 'r+'}
        File access mode, read-only ('r', default) or read-write ('r+')

    iline : int or segyio.TraceField
        Inline number field in the trace headers. Defaults to 189 as per the
        SEG-Y rev1 specification

    xline : int or segyio.TraceField
        Crossline number field in the trace headers. Defaults to 193 as per the
        SEG-Y rev1 specification

    strict : bool, optional
        Abort if a geometry cannot be inferred. Defaults to True.

    ignore_geometry : bool, optional
        Opt out on building geometry information, useful for e.g. shot
        organised files. Defaults to False.

    endian : {'big', 'msb', 'little', 'lsb'}
        File endianness, big/msb (default) or little/lsb

    Returns
    -------

    file : segyio.SegyFile
        An open segyio file handle

    Raises
    ------

    ValueError
        If the mode string contains 'w', as it would truncate the file

    Notes
    -----

    .. versionadded:: 1.1

    .. versionchanged:: 1.8
        endian argument

    When a file is opened non-strict, only raw traces access is allowed, and
    using modes such as ``iline`` raise an error.


    Examples
    --------

    Open a file in read-only mode:

    >>> with segyio.open(path, "r") as f:
    ...     print(f.ilines)
    ...
    [1, 2, 3, 4, 5]

    Open a file in read-write mode:

    >>> with segyio.open(path, "r+") as f:
    ...     f.trace = np.arange(100)

    Open two files at once:

    >>> with segyio.open(src_path) as src, segyio.open(dst_path, "r+") as dst:
    ...     dst.trace = src.trace # copy all traces from src to dst

    Open a file little-endian file:

    >>> with segyio.open(path, endian = 'little') as f:
    ...     f.trace[0]

    """

    if 'w' in mode:
        problem = 'w in mode would truncate the file'
        solution = 'use r+ to open in read-write'
        raise ValueError(', '.join((problem, solution)))

    endians = {
        'little': 256, # (1 << 8)
        'lsb': 256,
        'big': 0,
        'msb': 0,
    }

    if endian not in endians:
        problem = 'unknown endianness {}, expected one of: '
        opts = ' '.join(endians.keys())
        raise ValueError(problem.format(endian) + opts)

    from . import _segyio
    fd = _segyio.segyiofd(str(filename), mode, endians[endian])
    fd.segyopen()
    metrics = fd.metrics()

    f = segyio.SegyFile(fd,
            filename = str(filename),
            mode = mode,
            iline = iline,
            xline = xline,
            endian = endian,
    )

    try:
        dt = segyio.tools.dt(f, fallback_dt = 4000.0) / 1000.0
        t0 = f.header[0][segyio.TraceField.DelayRecordingTime]
        samples = metrics['samplecount']
        f._samples = (numpy.arange(samples) * dt) + t0

    except:
        f.close()
        raise

    if ignore_geometry:
        return f

    return infer_geometry(f, metrics, iline, xline, strict)



import warnings
try:
    from future_builtins import zip
    range = xrange
except (NameError, ImportError): pass

import numpy as np

from .gather import Gather, Groups
from .line import Line
from .trace import Trace, Header, Attributes, Text
from .field import Field

from .tracesortingformat import TraceSortingFormat



class SegyFile(object):
    """
    This class is not meant to be instantiated directly, but rather obtained
    from ``segyio.open`` or ``segyio.create``.
    """

    _unstructured_errmsg = "File opened in unstructured mode."

    def __init__(self, fd, filename, mode, iline=189,
                                           xline=193,
                                           endian='big',
                                           ):

        self._filename = filename
        self._mode = mode
        self._il = iline
        self._xl = xline

        # property value holders
        self._ilines = None
        self._xlines = None
        self._offsets = None
        self._samples = None
        self._sorting = None

        # private values
        self._iline_length = None
        self._iline_stride = None
        self._xline_length = None
        self._xline_stride = None

        self.xfd = fd
        metrics = self.xfd.metrics()
        self._fmt = metrics['format']
        self._tracecount = metrics['tracecount']
        self._ext_headers = metrics['ext_headers']

        try:
            self._dtype = np.dtype({
               -1: np.float32,
                1: np.float32,
                2: np.int32,
                3: np.int16,
                5: np.float32,
                6: np.float64,
                8: np.int8,
                9: np.int64,
                10: np.uint32,
                11: np.uint16,
                12: np.uint64,
                16: np.uint8,
            }[self._fmt])
        except KeyError:
            problem = 'Unknown trace value format {}'.format(self._fmt)
            solution = 'falling back to ibm float'
            warnings.warn(', '.join((problem, solution)))
            self._fmt = 1
            self._dtype = np.dtype(np.float32)

        self._trace = Trace(self.xfd,
                            self.dtype,
                            self.tracecount,
                            metrics['samplecount'],
                            self.readonly,
                           )
        self._header = Header(self)
        self._iline = None
        self._xline = None
        self._gather = None
        self.depth = None
        self.endian = endian

        super(SegyFile, self).__init__()

    def __str__(self):
        f = "SegyFile {}:".format(self._filename)

        if self.unstructured:
            il =  "  inlines: None"
            xl =  "  crosslines: None"
            of =  "  offsets: None"
        else:
            il =  "  inlines: {} [{}, {}]".format(len(self.ilines), self.ilines[0], self.ilines[-1])
            xl =  "  crosslines: {} [{}, {}]".format(len(self.xlines), self.xlines[0], self.xlines[-1])
            of =  "  offsets: {} [{}, {}]".format(len(self.offsets), self.offsets[0], self.offsets[-1])

        tr =  "  traces: {}".format(self.tracecount)
        sm =  "  samples: {}".format(self.samples)
        fmt = "  float representation: {}".format(self.format)

        props = [f, il, xl, tr, sm]

        if self.offsets is not None and len(self.offsets) > 1:
            props.append(of)

        props.append(fmt)
        return '\n'.join(props)


    def __repr__(self):
        return "SegyFile('{}', '{}', iline = {}, xline = {})".format(
                        self._filename, self._mode, self._il, self._xl)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def flush(self):
        """Flush the file

        Write the library buffers to disk, like C's ``fflush``. This method is
        mostly useful for testing.

        It is not necessary to call this method unless you want to observe your
        changes on-disk while the file is still open. The file will
        automatically be flushed for you if you use the `with` statement when
        your routine is completed.

        Notes
        -----

        .. versionadded:: 1.1

        .. warning::
            This is not guaranteed to actually write changes to disk, it only
            flushes the library buffers. Your kernel is free to defer writing
            changes to disk until a later time.

        Examples
        --------

        Flush:

        >>> with segyio.open(path) as f:
        ...     # write something to f
        ...     f.flush()

        """
        self.xfd.flush()

    def close(self):
        """Close the file

        This method is mostly useful for testing.

        It is not necessary to call this method if you're using the `with`
        statement, which will close the file for you. Calling methods on a
        previously-closed file will raise `IOError`.

        Notes
        -----

        .. versionadded:: 1.1


        """
        self.xfd.close()

    def mmap(self):
        """Memory map the file

        Memory map the file. This is an advanced feature for speed and
        optimization; however, it is no silver bullet. If your file is smaller
        than the memory available on your system this will likely result in
        faster reads and writes, especially for line modes. However, if the
        file is very large, or memory is very pressured, this optimization
        might cause overall system slowdowns. However, if you're opening the
        same file from many different instances of segyio then memory mapping
        may significantly reduce the memory pressure.

        If this call returns true, the file is memory mapped. If memory mapping
        was build-time disabled or is not available for your platform this call
        always return false. If the memory mapping is unsuccessful you can keep
        using segyio - reading and writing falls back on non-memory mapped
        features.

        Returns
        -------

        success : bool
            Returns True if the file was successfully memory mapped, False if
            not

        Notes
        -----

        .. versionadded:: 1.1


        Examples
        --------

        Memory map:

        >>> mapped = f.mmap()
        >>> if mapped: print( "File is memory mapped!" )
        File is memory mapped!
        >>> pass # keep using segyio as per usual
        >>> print( f.trace[10][7] )
        1.02548

        """
        return self.xfd.mmap()

    @property
    def dtype(self):
        """

        The data type object of the traces. This is the format most accurate
        and efficient to exchange with the underlying file, and the data type
        you will find the data traces.

        Returns
        -------

        dtype : numpy.dtype

        Notes
        -----

        .. versionadded:: 1.6

        """
        return self._dtype

    @property
    def sorting(self):
        """

        Inline or crossline sorting, or Falsey (None or 0) if unstructured.
        Returns
        -------

        sorting : int

        """
        return self._sorting

    @property
    def tracecount(self):
        """Number of traces in this file

        Equivalent to ``len(f.trace)``

        Returns
        -------

        count : int
            Number of traces in this file

        """
        return self._tracecount

    @property
    def samples(self):
        """
        Return the array of samples with appropriate intervals.

        Returns
        -------

        samples : numpy.ndarray of int

        Notes
        -----

        It holds that ``len(f.samples) == len(f.trace[0])``

        """

        return self._samples

    @property
    def offsets(self):
        """

        Return the array of offset names. For post-stack data, this array has a
        length of 1

        Returns
        -------

        offsets : numpy.ndarray of int

        """
        return self._offsets

    @property
    def ext_headers(self):
        """Extra text headers

        The number of extra text headers, given by the ``ExtendedHeaders``
        field in the binary header.

        Returns
        -------

        headers : int
            Number of extra text headers

        """

        return self._ext_headers

    @property
    def unstructured(self):
        """
        If the file is unstructured, sophisticated addressing modes that
        require the file to represent a proper cube won't work, and only raw
        data reading and writing is supported.

        Returns
        -------

        unstructured : bool
            ``True`` if this file is unstructured, ``False`` if not

        """
        return self.ilines is None

    @property
    def header(self):
        """
        Interact with segy in header mode

        Returns
        -------
        header : Header

        Notes
        -----
        .. versionadded:: 1.1
        """
        return self._header

    @header.setter
    def header(self, val):
        """headers macro assignment

        A convenient way for operating on all headers of a file is to use the
        default full-file range.  It will write headers 0, 1, ..., n, but uses
        the iteration specified by the right-hand side (i.e. can skip headers
        etc).

        If the right-hand-side headers are exhausted before all the destination
        file headers the writing will stop, i.e. not all all headers in the
        destination file will be written to.

        Examples
        --------
        Copy headers from file g to file f:

        >>> f.header = g.header

        Set offset field:

        >>> f.header = { TraceField.offset: 5 }

        Copy every 12th header from the file g to f's 0, 1, 2...:

        >>> f.header = g.header[::12]
        >>> f.header[0] == g.header[0]
        True
        >>> f.header[1] == g.header[12]
        True
        >>> f.header[2] == g.header[2]
        False
        >>> f.header[2] == g.header[24]
        True
        """
        self.header[:] = val

    def attributes(self, field):
        """File-wide attribute (header word) reading

        Lazily gather a single header word for every trace in the file. The
        array can be sliced, supports index lookup, and numpy-style
        list-of-indices.

        Parameters
        ----------

        field : int or segyio.TraceField
            field

        Returns
        -------

        attrs : Attributes
            A sliceable array_like of header words

        Notes
        -----

        .. versionadded:: 1.1

        """
        return Attributes(field, self.xfd, self.tracecount)

    @property
    def trace(self):
        """
        Interact with segy in trace mode.

        Returns
        -------
        trace : Trace

        Notes
        -----
        .. versionadded:: 1.1

        """

        return self._trace

    @trace.setter
    def trace(self, val):
        """traces macro assignment

        Convenient way for setting traces from 0, 1, ... n, based on the
        iterable set of traces on the right-hand-side.

        If the right-hand-side traces are exhausted before all the destination
        file traces the writing will stop, i.e. not all all traces in the
        destination file will be written.

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------
        Copy traces from file f to file g:

        >>> f.trace = g.trace

        Copy first half of the traces from g to f:

        >>> f.trace = g.trace[:len(g.trace)/2]

        Fill the file with one trace (filled with zeros):

        >>> tr = np.zeros(f.samples)
        >>> f.trace = itertools.repeat(tr)

        For advanced users: sometimes you want to load the entire segy file
        to memory and apply your own structural manipulations or operations
        on it. Some segy files are very large and may not fit, in which
        case this feature will break down. This is an optimisation feature;
        using it should generally be driven by measurements.

        Read the first 10 traces:

        >>> f.trace.raw[0:10]

        Read *all* traces to memory:

        >>> f.trace.raw[:]

        Read every other trace to memory:

        >>> f.trace.raw[::2]
        """
        self.trace[:] = val

    @property
    def ilines(self):
        """Inline labels

        The inline labels in this file, if structured, else None

        Returns
        -------

        inlines : array_like of int or None

        """
        return self._ilines

    @property
    def iline(self):
        """
        Interact with segy in inline mode

        Returns
        -------
        iline : Line or None

        Raises
        ------
        ValueError
            If the file is unstructured

        Notes
        -----
        .. versionadded:: 1.1
        """

        if self.unstructured:
            raise ValueError(self._unstructured_errmsg)

        if self._iline is not None:
            return self._iline

        self._iline = Line(self,
                           self.ilines,
                           self._iline_length,
                           self._iline_stride,
                           self.offsets,
                           'inline',
                          )
        return self._iline

    @iline.setter
    def iline(self, value):
        """inlines macro assignment

        Convenient way for setting inlines, from left-to-right as the inline
        numbers are specified in the file.ilines property, from an iterable
        set on the right-hand-side.

        If the right-hand-side inlines are exhausted before all the destination
        file inlines the writing will stop, i.e. not all all inlines in the
        destination file will be written.

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------
        Copy inlines from file f to file g:

        >>> f.iline = g.iline
        """
        self.iline[:] = value

    @property
    def xlines(self):
        """Crossline labels

        The crosslane labels in this file, if structured, else None

        Returns
        -------

        crosslines : array_like of int or None

        """
        return self._xlines

    @property
    def xline(self):
        """
        Interact with segy in crossline mode

        Returns
        -------
        xline : Line or None

        Raises
        ------
        ValueError
            If the file is unstructured

        Notes
        -----
        .. versionadded:: 1.1
        """
        if self.unstructured:
            raise ValueError(self._unstructured_errmsg)

        if self._xline is not None:
            return self._xline

        self._xline = Line(self,
                           self.xlines,
                           self._xline_length,
                           self._xline_stride,
                           self.offsets,
                           'crossline',
                          )
        return self._xline

    @xline.setter
    def xline(self, value):
        """crosslines macro assignment

        Convenient way for setting crosslines, from left-to-right as the inline
        numbers are specified in the file.ilines property, from an iterable set
        on the right-hand-side.

        If the right-hand-side inlines are exhausted before all the destination
        file inlines the writing will stop, i.e. not all all inlines in the
        destination file will be written.

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------
        Copy crosslines from file f to file g:

        >>> f.xline = g.xline
        """
        self.xline[:] = value

    @property
    def fast(self):
        """Access the 'fast' dimension

        This mode yields iline or xline mode, depending on which one is laid
        out `faster`, i.e. the line with linear disk layout. Use this mode if
        the inline/crossline distinction isn't as interesting as traversing in
        a fast manner (typically when you want to apply a function to the whole
        file, line-by-line).

        Returns
        -------
        fast : Line
            line addressing mode

        Notes
        -----
        .. versionadded:: 1.1
        """
        if self.sorting == TraceSortingFormat.INLINE_SORTING:
            return self.iline
        elif self.sorting == TraceSortingFormat.CROSSLINE_SORTING:
            return self.xline
        else:
            raise RuntimeError("Unknown sorting.")

    @property
    def slow(self):
        """Access the 'slow' dimension

        This mode yields iline or xline mode, depending on which one is laid
        out `slower`, i.e. the line with strided disk layout. Use this mode if
        the inline/crossline distinction isn't as interesting as traversing in
        the slower direction.

        Returns
        -------
        slow : Line
            line addressing mode

        Notes
        -----
        .. versionadded:: 1.1
        """
        if self.sorting == TraceSortingFormat.INLINE_SORTING:
            return self.xline
        elif self.sorting == TraceSortingFormat.CROSSLINE_SORTING:
            return self.iline
        else:
            raise RuntimeError("Unknown sorting.")

    @property
    def depth_slice(self):
        """
        Interact with segy in depth slice mode (fixed z-coordinate)

        Returns
        -------
        depth : Depth

        Notes
        -----
        .. versionadded:: 1.1

        .. versionchanged:: 1.7.1
            enabled for unstructured files
        """

        if self.depth is not None:
            return self.depth

        from .depth import Depth
        self.depth = Depth(self)
        return self.depth

    @depth_slice.setter
    def depth_slice(self, value):
        """depth macro assignment

        Convenient way for setting depth slices, from left-to-right as the depth slices
        numbers are specified in the file.depth_slice property, from an iterable
        set on the right-hand-side.

        If the right-hand-side depth slices are exhausted before all the destination
        file depth slices the writing will stop, i.e. not all all depth slices in the
        destination file will be written.

        Examples
        --------
        Copy depth slices from file f to file g:

        >>> f.depth_slice = g.depth_slice

        Copy first half of the depth slices from g to f:

        >>> f.depth_slice = g.depth_slice[:g.samples/2]]
        """
        self.depth_slice[:] = value

    @property
    def gather(self):
        """
        Interact with segy in gather mode

        Returns
        -------
        gather : Gather
        """
        if self.unstructured:
            raise ValueError(self._unstructured_errmsg)

        if self._gather is not None:
            return self._gather

        self._gather = Gather(self.trace, self.iline, self.xline, self.offsets)
        return self._gather

    @property
    def text(self):
        """Interact with segy in text mode

        This mode gives access to reading and writing functionality for textual
        headers.

        The primary data type is the python string. Reading textual headers is
        done with ``[]``, and writing is done via assignment. No additional
        structure is built around the textual header, so everything is treated
        as one long string without line breaks.

        Returns
        -------
        text : Text

        See also
        --------
        segyio.tools.wrap : line-wrap a text header

        Notes
        -----
        .. versionadded:: 1.1
        """
        return Text(self.xfd, self._ext_headers + 1)

    @property
    def bin(self):
        """
        Interact with segy in binary mode

        This mode gives access to reading and writing functionality for the
        binary header. Please note that using numeric binary offsets uses the
        offset numbers from the specification, i.e. the first field of the
        binary header starts at 3201, not 1. If you're using the enumerations
        this is handled for you.

        Returns
        -------
        binary : Field

        Notes
        -----
        .. versionadded:: 1.1
        """

        return Field.binary(self)

    @bin.setter
    def bin(self, value):
        """Update binary header

        Update a value or replace the binary header

        Parameters
        ----------

        value : dict_like
            dict_like, keys of int or segyio.BinField or segyio.su

        """
        self.bin.update(value)

    @property
    def format(self):
        d = {
           -2: "4-byte native big-endian float",
           -1: "4-byte native little-endian float",
            1: "4-byte IBM float",
            2: "4-byte signed integer",
            3: "2-byte signed integer",
            4: "4-byte fixed point with gain",
            5: "4-byte IEEE float",
            6: "8-byte IEEE float",
            7: "3-byte signed integer",
            8: "1-byte signed char",
            9: "8-byte signed integer",
            10: "4-byte unsigned integer",
            11: "2-byte unsigned integer",
            12: "8-byte unsigned integer",
            15: "3-byte unsigned integer",
            16: "1-byte unsigned char"
        }

        class fmt:
            def __int__(inner):
                return self._fmt

            def __str__(inner):
                if not self._fmt in d:
                    return "Unknown format"

                return d[self._fmt]

        return fmt()

    @property
    def readonly(self):
        """File is read-only

        Returns
        -------
        readonly : bool
            True if this file is read-only

        Notes
        -----
        .. versionadded:: 1.6
        """

        return '+' not in self._mode


    def interpret(self, ilines, xlines, offsets=None, sorting=TraceSortingFormat.INLINE_SORTING):

        """ (Re-)interpret structure on top of a file

        (Re-)interpret the structure of the file given the new sorting, ilines,
        xlines and offset indices. Note that file itself is not changed in any
        way, it is only segyio's interpretation of the file that changes. It's
        a way of telling segyio that a file is laid out in a particular way,
        even though the header fields say otherwise.

        `interpret` expect that the ilines-, xlines- and offsets-indices are
        unique. It also expect the dimensions of ilines, xlines and offset to
        match the tracecount.

        Parameters
        ----------
        f : SegyFile
        ilines : array_like
            ilines indices in new structure
        xlines : array_like
            xlines indices in new structure
        offsets : array_like
            offset indices in new structure
        sorting : int, string or TraceSortingFormat
            Sorting in new structure

        Notes
        -----

        .. versionadded:: 1.8

        Examples
        --------
        (Re)interpret the structure of the file:

        >>> ilines = [10, 11, 12, 13]
        >>> xlines = [20, 21, 22, 23, 24]
        >>> with segyio.open(file, ignore_geometry=True) as f:
        ... f.interpret(ilines, xlines)
        """

        valid_sortings = {
            1           : TraceSortingFormat.CROSSLINE_SORTING,
            2           : TraceSortingFormat.INLINE_SORTING,
            'iline'     : TraceSortingFormat.INLINE_SORTING,
            'inline'    : TraceSortingFormat.INLINE_SORTING,
            'xl'        : TraceSortingFormat.CROSSLINE_SORTING,
            'crossline' : TraceSortingFormat.CROSSLINE_SORTING,
            TraceSortingFormat.INLINE_SORTING    : TraceSortingFormat.INLINE_SORTING,
            TraceSortingFormat.CROSSLINE_SORTING : TraceSortingFormat.CROSSLINE_SORTING,
        }

        if sorting not in valid_sortings:
            error = "Invalid sorting"
            solution = "valid sorting options are: {}".format(valid_sortings.keys())
            raise ValueError('{}, {}'.format(error, solution))

        if offsets is None:
            offsets = np.arange(1)

        ilines  = np.copy(np.asarray(ilines,  dtype=np.intc))
        xlines  = np.copy(np.asarray(xlines,  dtype=np.intc))
        offsets = np.copy(np.asarray(offsets, dtype=np.intc))

        if np.unique(ilines).size != ilines.size:
            error = "Inlines inconsistent"
            solution = "expect all inlines to be unique"
            raise ValueError("{}, {}".format(error, solution))

        if np.unique(xlines).size != xlines.size:
            error = "Crosslines inconsistent"
            solution = "expect all crosslines to be unique"
            raise ValueError("{}, {}".format(error, solution))

        if np.unique(offsets).size != offsets.size:
            error = "Offsets inconsistent"
            solution = "expect all offsets to be unique"
            raise ValueError("{}, {}".format(error, solution))

        if ilines.size * xlines.size * offsets.size != self.tracecount:
            error = ("Invalid dimensions, ilines ({}) * xlines ({}) * offsets "
                     "({}) should match the number of traces ({})").format(ilines.size,
                                                                           xlines.size,
                                                                           offsets.size,
                                                                           self.tracecount)
            raise ValueError(error)

        from . import _segyio

        line_metrics = _segyio.line_metrics(sorting,
                                            self.tracecount,
                                            ilines.size,
                                            xlines.size,
                                            offsets.size)

        self._iline_length = line_metrics['iline_length']
        self._iline_stride = line_metrics['iline_stride']

        self._xline_length = line_metrics['xline_length']
        self._xline_stride = line_metrics['xline_stride']

        self._sorting = sorting
        self._offsets = offsets
        self._ilines = ilines
        self._xlines = xlines

        return self

    def group(self, word):
        """Get groups of traces matching a key

        This feature is **experimental**, and there are no guarantees code
        using this will work in the future.

        Walks the headers and groups traces into buckets, where all traces in a
        bucket have the same value for the given set of words. It is
        particularly useful for pre-stack files, gathering traces belonging to
        the same gather or shot.

        Parameters
        ----------
        word : segyio.TraceField or iterable of segyio.TraceField
            The set of words belonging to the same bucket

        Returns
        -------
        groups : segyio.gather.Groups

        Notes
        -----
        This feature is **experimental**, but you are encouraged to try it out.
        Bug reports and suggestions for improvement are welcomed, but no
        guarantees are made that the interface will remain as it is in the
        future.
        """
        msg = 'group is experimental and may change in the future'
        warnings.warn(msg, FutureWarning)
        return Groups(self.trace, self.header, word)


class spec(object):
    def __init__(self):
        self.iline = 189
        self.ilines = None
        self.xline = 193
        self.xlines = None
        self.offsets = [1]
        self.samples = None
        self.ext_headers = 0
        self.format = None
        self.sorting = None
        self.endian = 'big'



from ..open import infer_geometry
from ..segy import SegyFile
from . import words

import numpy

class sufile(SegyFile):
    def __init__(self, *args, **kwargs):
        super(sufile, self).__init__(*args, **kwargs)

    @property
    def text(self):
        raise NotImplementedError

    @property
    def bin(self):
        raise NotImplementedError

    @bin.setter
    def bin(self, _):
        raise NotImplementedError

def open(filename, mode = 'r', iline = 189,
                               xline = 193,
                               strict = True,
                               ignore_geometry = False,
                               endian = 'big' ):
    """Open a seismic unix file.

    Behaves identically to open(), except it expects the seismic unix format,
    not SEG-Y.

    Parameters
    ----------
    filename : str
        Path to file to open

    mode : {'r', 'r+'}
        File access mode, read-only ('r', default) or read-write ('r+')

    iline : int or segyio.TraceField
        Inline number field in the trace headers. Defaults to 189 as per the
        SEG-Y rev1 specification

    xline : int or segyio.TraceField
        Crossline number field in the trace headers. Defaults to 193 as per the
        SEG-Y rev1 specification

    strict : bool, optional
        Abort if a geometry cannot be inferred. Defaults to True.

    ignore_geometry : bool, optional
        Opt out on building geometry information, useful for e.g. shot
        organised files. Defaults to False.

    endian : {'big', 'msb', 'little', 'lsb'}
        File endianness, big/msb (default) or little/lsb

    Returns
    -------
    file : segyio.su.file
        An open seismic unix file handle

    Raises
    ------
    ValueError
        If the mode string contains 'w', as it would truncate the file

    See also
    --------
    segyio.open : SEG-Y open

    Notes
    -----
    .. versionadded:: 1.8
    """

    if 'w' in mode:
        problem = 'w in mode would truncate the file'
        solution = 'use r+ to open in read-write'
        raise ValueError(', '.join((problem, solution)))

    endians = {
        'little': 256, # (1 << 8)
        'lsb': 256,
        'big': 0,
        'msb': 0,
    }

    if endian not in endians:
        problem = 'unknown endianness, must be one of: '
        candidates = ' '.join(endians.keys())
        raise ValueError(problem + candidates)

    from .. import _segyio
    fd = _segyio.segyiofd(str(filename), mode, endians[endian])
    fd.suopen()
    metrics = fd.metrics()

    f = sufile(
        fd,
        filename = str(filename),
        mode = mode,
        iline = iline,
        xline = xline,
    )

    h0 = f.header[0]

    dt = h0[words.dt] / 1000.0
    t0 = h0[words.delrt]
    samples = metrics['samplecount']
    f._samples = (numpy.arange(samples) * dt) + t0

    if ignore_geometry:
        return f

    return infer_geometry(f, metrics, iline, xline, strict)



"""segyio

Welcome to segyio. For help, examples and reference, type ``help(function)`` in
your favourite python interpreter, or ``pydoc function`` in the unix console.

The segyio library attempts to be easy to use efficently for prototyping and
interaction with possibly large segy files. File reading and writing is
streaming, with large file support out of the box and without hassle. For a
quick start on reading files, type ``help(segyio.open)``.

An open segy file is interacted with in modes. For a reference with examples,
please type ``help(segyio.segy)``, look at the online documentation at
segyio.readthedocs.io, or run ``help()`` on the object returned by
``segyio.open``.. For documentation on individual modes, please
refer to the individual modes with ``help(f.[mode])``, where ``f`` is an open
file handle.

The available modes are:
    * text, for textual headers including extended headers
    * bin, for the binary header
    * header, for the trace headers
    * trace, for trace data
    * iline, for inline biased operations
    * xline, for crossline biased operations
    * depth_slice, for depth biased operations
    * gather, for gather/intersaction biased operations

The primary data type is the numpy.ndarray. All examples use ``np`` for the
numpy namespace. That means that any function that returns a trace, a set of
samples or even full lines, returns a numpy.ndarray. This enables quick and
easy mathematical operations on the data you care about.

Segyio is designed to blend into regular python code, so python concepts that
map to segy operations are written to behave similarly. That means that
sequences of data support list lookup, slicing (``f.trace[0:10:2]``), ``for x
in`` etc. Please refer to the individual modes' documentation for a more
extensive set of examples.

For all slicing operations that segyio provides the underlying buffer is
reused, so if you want to keep the data between iterations it is necessary to
manually copy the data, e.g. ``numpy.copy()``. Please refer to the examples.
"""


class Enum(object):
    def __init__(self, enum_value):
        super(Enum, self).__init__()
        self._value = int(enum_value)

    def __int__(self):
        return int(self._value)

    def __str__(self):
        for k, v in self.__class__.__dict__.items():
            if isinstance(v, int) and self._value == v:
                return k
        return "Unknown Enum"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(self._value)

    def __eq__(self, other):
        try:
            o = int(other)
        except ValueError:
            return super(Enum, self).__eq__(other)
        else:
            return self._value == o

    def __ne__(self, other):
        return not self == other

    @classmethod
    def enums(cls):
        result = []
        for v in cls.__dict__.values():
            if isinstance(v, int):
                result.append(cls(v))

        return sorted(result, key=int)


from .binfield import BinField
from .segysampleformat import SegySampleFormat
from .tracesortingformat import TraceSortingFormat
from .tracefield import TraceField
from . import su
from .open import open
from .create import create
from .segy import SegyFile, spec
from .tools import dt, sample_indexes, create_text_header, native
from .tools import collect, cube

try:
    from .version import version as __version__
except ImportError:
    __version__ = '0.0.0'

__copyright__  = 'Copyright 2016, Statoil ASA'
__license__    = 'GNU Lesser General Public License version 3'
__status__     = 'Production'



from . import Enum

class BinField(Enum):
    """Trace header field enumerator

    See also
    -------
    segyio.su : Seismic unix aliases for header fields
    """

    JobID = 3201
    LineNumber = 3205
    ReelNumber = 3209
    Traces = 3213
    AuxTraces = 3215
    Interval = 3217
    IntervalOriginal = 3219
    Samples = 3221
    SamplesOriginal = 3223
    Format = 3225
    EnsembleFold = 3227
    SortingCode = 3229
    VerticalSum = 3231
    SweepFrequencyStart = 3233
    SweepFrequencyEnd = 3235
    SweepLength = 3237
    Sweep = 3239
    SweepChannel = 3241
    SweepTaperStart = 3243
    SweepTaperEnd = 3245
    Taper = 3247
    CorrelatedTraces = 3249
    BinaryGainRecovery = 3251
    AmplitudeRecovery = 3253
    MeasurementSystem = 3255
    ImpulseSignalPolarity = 3257
    VibratoryPolarity = 3259
    ExtTraces = 3261
    ExtAuxTraces = 3265
    ExtSamples = 3269
    ExtSamplesOriginal = 3289
    ExtEnsembleFold = 3293
    Unassigned1 = 3261
    SEGYRevision = 3501
    SEGYRevisionMinor = 3502
    TraceFlag = 3503
    ExtendedHeaders = 3505
    Unassigned2 = 3507

keys = {
    'JobID'                 : 3201,
    'LineNumber'            : 3205,
    'ReelNumber'            : 3209,
    'Traces'                : 3213,
    'AuxTraces'             : 3215,
    'Interval'              : 3217,
    'IntervalOriginal'      : 3219,
    'Samples'               : 3221,
    'SamplesOriginal'       : 3223,
    'Format'                : 3225,
    'EnsembleFold'          : 3227,
    'SortingCode'           : 3229,
    'VerticalSum'           : 3231,
    'SweepFrequencyStart'   : 3233,
    'SweepFrequencyEnd'     : 3235,
    'SweepLength'           : 3237,
    'Sweep'                 : 3239,
    'SweepChannel'          : 3241,
    'SweepTaperStart'       : 3243,
    'SweepTaperEnd'         : 3245,
    'Taper'                 : 3247,
    'CorrelatedTraces'      : 3249,
    'BinaryGainRecovery'    : 3251,
    'AmplitudeRecovery'     : 3253,
    'MeasurementSystem'     : 3255,
    'ImpulseSignalPolarity' : 3257,
    'VibratoryPolarity'     : 3259,
    'ExtTraces'             : 3261,
    'ExtAuxTraces'          : 3265,
    'ExtSamples'            : 3269,
    'ExtSamplesOriginal'    : 3289,
    'ExtEnsembleFold'       : 3293,
    'Unassigned1'           : 3301,
    'SEGYRevision'          : 3501,
    'SEGYRevisionMinor'     : 3502,
    'TraceFlag'             : 3503,
    'ExtendedHeaders'       : 3505,
    'Unassigned2'           : 3507,
}



try:
    from collections.abc import Mapping # noqa
except ImportError:
    from collections import Mapping # noqa

import itertools
try: from future_builtins import zip
except ImportError: pass
import numpy as np

from .utils import castarray

# in order to support [:end] syntax, we must make sure
# start has a non-None value. lineno.indices() would set it
# to 0, but we don't know if that's a reasonable value or
# not. If start is None we set it to the first line
def sanitize_slice(s, source):
    if all((s.start, s.stop, s.step)):
        return s

    start, stop, step = s.start, s.stop, s.step
    increasing = step is None or step > 0

    if start is None:
        start = min(source) if increasing else max(source)

    if stop is None:
        stop = max(source) + 1 if increasing else min(source) - 1

    return slice(start, stop, step)

class Line(Mapping):
    """
    The Line implements the dict interface, with a fixed set of int_like keys,
    the line numbers/labels. Data is read lazily from disk, so iteration does
    not consume much memory, and are returned as numpy.ndarrays.

    It provides a convenient interface for reading data in a cartesian grid
    system, provided one exists and is detectable by segyio.

    Lines can be accessed individually or with slices, and writing is done via
    assignment. Note that accessing lines uses the line numbers, not their
    position, so if a files has lines [2400..2500], accessing line [0..100]
    will be an error. Note that since each line is returned as a numpy.ndarray,
    meaning accessing the intersections of the inline and crossline is
    0-indexed - orthogonal labels are not preserved.

    Additionally, the line has a concept of offsets, which is useful when
    dealing with prestack files. Offsets are accessed via sub indexing, meaning
    iline[10, 4] will give you line 10 at offset 4.  Please note that offset,
    like lines, are accessed via their labels, not their indices. If your file
    has the offsets [150, 250, 350, 450] and the lines [2400..2500], you can
    access the third offset with [2403, 350]. Please refer to the examples for
    more details. If no offset is specified, segyio will give you the first.

    Notes
    -----
    .. versionadded:: 1.1

    .. versionchanged:: 1.6
        common dict operations (Mapping)
    """

    def __init__(self, filehandle, labels, length, stride, offsets, name):
        self.filehandle = filehandle.xfd
        self.lines = labels
        self.length = length
        self.stride = stride
        self.shape = (length, len(filehandle.samples))
        self.dtype = filehandle.dtype

        # pre-compute all line beginnings
        from ._segyio import fread_trace0
        self.heads = {
            label: fread_trace0(label,
                                length,
                                stride,
                                len(offsets),
                                labels,
                                name)
            for label in labels
        }

        self.offsets = { x: i for i, x in enumerate(offsets) }
        self.default_offset = offsets[0]

    def ranges(self, index, offset):
        if not isinstance(index, slice):
            index = slice(index, index + 1)

        if not isinstance(offset, slice):
            offset = slice(offset, offset + 1)

        index  = sanitize_slice(index, self.heads.keys())
        offset = sanitize_slice(offset, self.offsets.keys())
        irange = range(*index.indices(max(self.heads.keys()) + 1))
        orange = range(*offset.indices(max(self.offsets.keys()) + 1))
        irange = filter(self.heads.__contains__, irange)
        orange = filter(self.offsets.__contains__, orange)
        # offset-range is used in inner loops, so make it a list for
        # reusability. offsets are usually few, so no real punishment by using
        # non-generators here
        return irange, list(orange)

    def __getitem__(self, index):
        """line[i] or line[i, o]

        The line `i`, or the line `i` at a specific offset `o`. ``line[i]``
        returns a numpy array, and changes to this array will *not* be
        reflected on disk.

        The `i` and `o` are *keys*, and should correspond to the line- and
        offset labels in your file, and in the `ilines`, `xlines`, and
        `offsets` attributes.

        Slices can contain lines and offsets not in the file, and like with
        list slicing, these are handled gracefully and ignored.

        When `i` or `o` is a slice, a generator of numpy arrays is returned. If
        the slice is defaulted (:), segyio knows enough about the structure to
        give you all of the respective labels.

        When both `i` and `o` are slices, only one generator is returned, and
        the lines are yielded offsets-first, roughly equivalent to the double
        for loop::

            >>> for line in lines:
            ...     for off in offsets:
            ...         yield line[line, off]
            ...

        Parameters
        ----------
        i : int or slice
        o : int or slice

        Returns
        -------
        line : numpy.ndarray of dtype or generator of numpy.ndarray of dtype

        Raises
        ------
        KeyError
            If `i` or `o` don't exist

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------

        Read an inline:

        >>> x = line[2400]

        Copy every inline into a list:

        >>> l = [numpy.copy(x) for x in iline[:]]

        Numpy operations on every other inline:

        >>> for line in line[::2]:
        ...     line = line * 2
        ...     avg = np.average(line)

        Read lines up to 2430:

        >>> for line in line[:2430]:
        ...     line.mean()

        Copy all lines at all offsets:

        >>> l = [numpy.copy(x) for x in line[:,:]]

        Copy all offsets of a line:

        >>> x = numpy.copy(iline[10,:])

        Copy all lines at a fixed offset:

        >>> x = numpy.copy(iline[:, 120])

        Copy every other line and offset:

        >>> map(numpy.copy, line[::2, ::2])

        Copy all offsets [200, 250, 300, 350, ...] in the range [200, 800) for
        all lines [2420,2460):

        >>> l = [numpy.copy(x) for x in line[2420:2460, 200:800:50]]
        """

        offset = self.default_offset
        try: index, offset = index
        except TypeError: pass

        # prioritise the code path that's potentially in loops externally
        try:
            head = self.heads[index] + self.offsets[offset]
        except TypeError:
            # index is either unhashable (because it's a slice), or offset is a
            # slice.
            pass
        else:
            return self.filehandle.getline(head,
                                           self.length,
                                           self.stride,
                                           len(self.offsets),
                                           np.empty(self.shape, dtype=self.dtype),
                                          )

        # at this point, either offset or index is a slice (or proper
        # type-error), so we're definitely making a generator. make them both
        # slices to unify all code paths
        irange, orange = self.ranges(index, offset)

        def gen():
            x = np.empty(self.shape, dtype=self.dtype)
            y = np.copy(x)

            # only fetch lines that exist. the slice can generate both offsets
            # and line numbers that don't exist, so filter out misses before
            # they happen
            for line in irange:
                for off in orange:
                    head = self.heads[line] + self.offsets[off]
                    self.filehandle.getline(head,
                                            self.length,
                                            self.stride,
                                            len(self.offsets),
                                            y,
                                           )
                    y, x = x, y
                    yield x

        return gen()

    def __setitem__(self, index, val):
        """line[i] = val or line[i, o] = val

        Follows the same rules for indexing and slicing as ``line[i]``.

        In either case, if the `val` iterable is exhausted before the line(s),
        assignment stops with whatever is written so far. If `val` is longer
        than an individual line, it's essentially truncated.

        Parameters
        ----------
        i       : int or slice
        offset  : int or slice
        val     : array_like

        Raises
        ------
        KeyError
            If `i` or `o` don't exist

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------
        Copy a full line:

        >>> line[2400] = other[2834]

        Copy first half of the inlines from g to f:

        >>> line[:] = other[:labels[len(labels) / 2]]

        Copy every other line consecutively:

        >>> line[:] = other[::2]

        Copy every third offset:

        >>> line[:,:] = other[:,::3]

        Copy a line into a set line and offset:

        >>> line[12, 200] = other[21]
        """

        offset = self.default_offset
        try: index, offset = index
        except TypeError: pass

        try: head = self.heads[index] + self.offsets[offset]
        except TypeError: pass
        else:
            return self.filehandle.putline(head,
                                           self.length,
                                           self.stride,
                                           len(self.offsets),
                                           index,
                                           offset,
                                           castarray(val, dtype = self.dtype),
                                          )

        irange, orange = self.ranges(index, offset)

        val = iter(val)
        for line in irange:
            for off in orange:
                head = self.heads[line] + self.offsets[off]
                try: self.filehandle.putline(head,
                                            self.length,
                                            self.stride,
                                            len(self.offsets),
                                            line,
                                            off,
                                            next(val),
                                           )
                except StopIteration: return

    # can't rely on most Mapping default implementations of
    # dict-like, because iter() does not yield keys for this class, it gives
    # the lines themselves. that violates some assumptions (but segyio's always
    # worked that way), and it's the more natural behaviour for segyio, so it's
    # acceptible. additionally, the default implementations would be very slow
    # and ineffective because they assume __getitem__ is sufficiently cheap,
    # but it isn't here since it involves a disk operation
    def __len__(self):
        """x.__len__() <==> len(x)"""
        return len(self.heads)

    def __iter__(self):
        """x.__iter__() <==> iter(x)"""
        return self[:]

    def __contains__(self, key):
        """x.__contains__(y) <==> y in x"""
        return key in self.heads

    def keys(self):
        """D.keys() -> a set-like object providing a view on D's keys"""
        return sorted(self.heads.keys())

    def values(self):
        """D.values() -> generator of D's values"""
        return self[:]

    def items(self):
        """D.values() -> generator of D's (key,values), as 2-tuples"""
        return zip(self.keys(), self[:])

class HeaderLine(Line):
    """
    The Line implements the dict interface, with a fixed set of int_like keys,
    the line numbers/labels. The values are iterables of Field objects.

    Notes
    -----
    .. versionadded:: 1.1

    .. versionchanged:: 1.6
        common dict operations (Mapping)
    """
    # a lot of implementation details are shared between reading data traces
    # line-by-line and trace headers line-by-line, so (ab)use inheritance for
    # __len__, keys() etc., however, the __getitem__ is  way different and is re-implemented

    def __init__(self, header, base, direction):
        super(HeaderLine, self).__init__(header.segy,
                                          base.lines,
                                          base.length,
                                          base.stride,
                                          sorted(base.offsets.keys()),
                                          'header.' + direction,
                                         )
        self.header = header

    def __getitem__(self, index):
        """line[i] or line[i, o]

        The line `i`, or the line `i` at a specific offset `o`. ``line[i]``
        returns an iterable of `Field` objects, and changes to these *will* be
        reflected on disk.

        The `i` and `o` are *keys*, and should correspond to the line- and
        offset labels in your file, and in the `ilines`, `xlines`, and
        `offsets` attributes.

        Slices can contain lines and offsets not in the file, and like with
        list slicing, these are handled gracefully and ignored.

        When `i` or `o` is a slice, a generator of iterables of headers are
        returned.

        When both `i` and `o` are slices, one generator is returned for the
        product `i` and `o`, and the lines are yielded offsets-first, roughly
        equivalent to the double for loop::

            >>> for line in lines:
            ...     for off in offsets:
            ...         yield line[line, off]
            ...

        Parameters
        ----------

        i : int or slice
        o : int or slice

        Returns
        -------
        line : iterable of Field or generator of iterator of Field

        Raises
        ------
        KeyError
            If `i` or `o` don't exist

        Notes
        -----
        .. versionadded:: 1.1

        """
        offset = self.default_offset
        try: index, offset = index
        except TypeError: pass

        try:
            start = self.heads[index] + self.offsets[offset]
        except TypeError:
            # index is either unhashable (because it's a slice), or offset is a
            # slice.
            pass

        else:
            step = self.stride * len(self.offsets)
            stop = start + step * self.length
            return self.header[start:stop:step]

        def gen():
            irange, orange = self.ranges(index, offset)
            for line in irange:
                for off in orange:
                    yield self[line, off]

        return gen()

    def __setitem__(self, index, val):
        """line[i] = val or line[i, o] = val

        Follows the same rules for indexing and slicing as ``line[i]``. If `i`
        is an int, and `val` is a dict or Field, that value is replicated and
        assigned to every trace header in the line, otherwise it's treated as
        an iterable, and each trace in the line is assigned the ``next()``
        yielded value.

        If `i` or `o` is a slice, `val` must be an iterable.

        In either case, if the `val` iterable is exhausted before the line(s),
        assignment stops with whatever is written so far.

        Parameters
        ----------
        i       : int or slice
        offset  : int or slice
        val     : dict_like or iterable of dict_like

        Raises
        ------
        KeyError
            If `i` or `o` don't exist

        Notes
        -----
        .. versionadded:: 1.1

        Examples
        --------
        Rename the iline 3 to 4:

        >>> line[3] = { TraceField.INLINE_3D: 4 }
        >>> # please note that rewriting the header won't update the
        >>> # file's interpretation of the file until you reload it, so
        >>> # the new iline 4 will be considered iline 3 until the file
        >>> # is reloaded

        Set offset line 3 offset 3 to 5:

        >>> line[3, 3] = { TraceField.offset: 5 }
        """
        offset = self.default_offset
        try: index, offset = index
        except TypeError: pass

        try: start = self.heads[index] + self.offsets[offset]
        except TypeError: pass
        else:
            step = self.stride * len(self.offsets)
            stop  = start + step * self.length
            self.header[start:stop:step] = val
            return

        # if this is a dict-like, just repeat it
        if hasattr(val, 'keys'):
            val = itertools.repeat(val)

        irange, orange = self.ranges(index, offset)
        val = iter(val)
        for line in irange:
            for off in orange:
                try:
                    self[line, off] = next(val)
                except StopIteration:
                    return



from . import Enum

class SegySampleFormat(Enum):
    IBM_FLOAT_4_BYTE = 1
    SIGNED_INTEGER_4_BYTE = 2
    SIGNED_SHORT_2_BYTE = 3
    FIXED_POINT_WITH_GAIN_4_BYTE = 4
    IEEE_FLOAT_4_BYTE = 5
    IEEE_FLOAT_8_BYTE = 6
    SIGNED_CHAR_3_BYTE = 7
    SIGNED_CHAR_1_BYTE = 8
    SIGNED_INTEGER_8_BYTE = 9
    UNSIGNED_INTEGER_4_BYTE = 10
    UNSIGNED_SHORT_2_BYTE = 11
    UNSIGNED_INTEGER_8_BYTE = 12
    UNSIGNED_INTEGER_3_BYTE = 15
    UNSIGNED_CHAR_1_BYTE = 16
    NOT_IN_USE_1 = 19
    NOT_IN_USE_2 = 20



"""Seismic Unix aliases

Seismic Unix style aliases for binary and trace header fields.

Notes
-----

.. versionadded: 1.1

Seismic Unix does not have names for all possible fields, as it came
around during an early revision of SEG-Y, and names for these fields
are created by segyio in their absence. If Seismic Unix starts
providing names for these fields, they will be added to these alies,
and in conflicts take presedence after some time. If there are no
conflicts, segyio names are considered for deprecation on a
case-by-case basis, but will most likely be supported along with the
Seismic Unix name.
"""

from ..tracefield import TraceField as tf
from ..binfield import BinField as bf

# the wonky #: before every line is so that sphinx' autodoc picks it up, and includes the constants in the documentation

#:
tracl  = tf.TRACE_SEQUENCE_LINE
#:
tracr  = tf.TRACE_SEQUENCE_FILE
#:
fldr   = tf.FieldRecord
#:
tracf  = tf.TraceNumber
#:
ep     = tf.EnergySourcePoint
#:
cdp    = tf.CDP
#:
cdpt   = tf.CDP_TRACE
#:
trid   = tf.TraceIdentificationCode
#:
nvs    = tf.NSummedTraces
#:
nhs    = tf.NStackedTraces
#:
duse   = tf.DataUse
#:
offset = tf.offset
#:
gelev  = tf.ReceiverGroupElevation
#:
selev  = tf.SourceSurfaceElevation
#:
sdepth = tf.SourceDepth
#:
gdel   = tf.ReceiverDatumElevation
#:
sdel   = tf.SourceDatumElevation
#:
swdep  = tf.SourceWaterDepth
#:
gwdep  = tf.GroupWaterDepth
#:
scalel = tf.ElevationScalar
#:
scalco = tf.SourceGroupScalar
#:
sx     = tf.SourceX
#:
sy     = tf.SourceY
#:
gx     = tf.GroupX
#:
gy     = tf.GroupY
#:
counit = tf.CoordinateUnits
#:
wevel  = tf.WeatheringVelocity
#:
swevel = tf.SubWeatheringVelocity
#:
sut    = tf.SourceUpholeTime
#:
gut    = tf.GroupUpholeTime
#:
sstat  = tf.SourceStaticCorrection
#:
gstat  = tf.GroupStaticCorrection
#:
tstat  = tf.TotalStaticApplied
#:
laga   = tf.LagTimeA
#:
lagb   = tf.LagTimeB
#:
delrt  = tf.DelayRecordingTime
#:
muts   = tf.MuteTimeStart
#:
mute   = tf.MuteTimeEND
#:
ns     = tf.TRACE_SAMPLE_COUNT
#:
dt     = tf.TRACE_SAMPLE_INTERVAL
#:
gain   = tf.GainType
#:
igc    = tf.InstrumentGainConstant
#:
igi    = tf.InstrumentInitialGain
#:
corr   = tf.Correlated
#:
sfs    = tf.SweepFrequencyStart
#:
sfe    = tf.SweepFrequencyEnd
#:
slen   = tf.SweepLength
#:
styp   = tf.SweepType
#:
stat   = tf.SweepTraceTaperLengthStart
#:
stae   = tf.SweepTraceTaperLengthEnd
#:
tatyp  = tf.TaperType
#:
afilf  = tf.AliasFilterFrequency
#:
afils  = tf.AliasFilterSlope
#:
nofilf = tf.NotchFilterFrequency
#:
nofils = tf.NotchFilterSlope
#:
lcf    = tf.LowCutFrequency
#:
hcf    = tf.HighCutFrequency
#:
lcs    = tf.LowCutSlope
#:
hcs    = tf.HighCutSlope
#:
year   = tf.YearDataRecorded
#:
day    = tf.DayOfYear
#:
hour   = tf.HourOfDay
#:
minute = tf.MinuteOfHour
#:
sec    = tf.SecondOfMinute
#:
timbas = tf.TimeBaseCode
#:
trwf   = tf.TraceWeightingFactor
#:
grnors = tf.GeophoneGroupNumberRoll1
#:
grnofr = tf.GeophoneGroupNumberFirstTraceOrigField
#:
grnlof = tf.GeophoneGroupNumberLastTraceOrigField
#:
gaps   = tf.GapSize
#:
otrav  = tf.OverTravel
#:
cdpx   = tf.CDP_X
#:
cdpy   = tf.CDP_Y
#:
iline  = tf.INLINE_3D
#:
xline  = tf.CROSSLINE_3D
#:
sp     = tf.ShotPoint
#:
scalsp = tf.ShotPointScalar
#:
trunit = tf.TraceValueMeasurementUnit
#:
tdcm   = tf.TransductionConstantMantissa
#:
tdcp   = tf.TransductionConstantPower
#:
tdunit = tf.TransductionUnit
#:
triden = tf.TraceIdentifier
#:
sctrh  = tf.ScalarTraceHeader
#:
stype  = tf.SourceType
#:
sedm   = tf.SourceEnergyDirectionMantissa
#:
sede   = tf.SourceEnergyDirectionExponent
#:
smm    = tf.SourceMeasurementMantissa
#:
sme    = tf.SourceMeasurementExponent
#:
smunit = tf.SourceMeasurementUnit
#:
uint1  = tf.UnassignedInt1
#:
uint2  = tf.UnassignedInt2

# binary
#:
jobid  = bf.JobID
#:
lino   = bf.LineNumber
#:
reno   = bf.ReelNumber
#:
ntrpr  = bf.Traces
#:
nart   = bf.AuxTraces
#:
hdt    = bf.Interval
#:
dto    = bf.IntervalOriginal
#:
hns    = bf.Samples
#:
nso    = bf.SamplesOriginal
#:
format = bf.Format
#:
fold   = bf.EnsembleFold
#:
tsort  = bf.SortingCode
#:
vscode = bf.VerticalSum
#:
hsfs   = bf.SweepFrequencyStart
#:
hsfe   = bf.SweepFrequencyEnd
#:
hslen  = bf.SweepLength
#:
hstyp  = bf.Sweep
#:
schn   = bf.SweepChannel
#:
hstas  = bf.SweepTaperStart
#:
hstae  = bf.SweepTaperEnd
#:
htatyp = bf.Taper
#:
hcorr  = bf.CorrelatedTraces
#:
bgrcv  = bf.BinaryGainRecovery
#:
rcvm   = bf.AmplitudeRecovery
#:
mfeet  = bf.MeasurementSystem
#:
polyt  = bf.ImpulseSignalPolarity
#:
vpol   = bf.VibratoryPolarity
#:
unas1  = bf.Unassigned1
#:
extntrpr  = bf.ExtTraces
#:
extnart   = bf.ExtAuxTraces
#:
exthns    = bf.ExtSamples
#:
extnso    = bf.ExtSamplesOriginal
#:
extfold   = bf.ExtEnsembleFold
#:
rev    = bf.SEGYRevision
#:
revmin = bf.SEGYRevisionMinor
#:
trflag = bf.TraceFlag
#:
exth   = bf.ExtendedHeaders
#:
unas2  = bf.Unassigned2



try:
    from collections.abc import Sequence # noqa
except ImportError:
    from collections import Sequence # noqa

import numpy as np
try: from future_builtins import zip
except ImportError: pass
from .trace import Sequence
from .utils import castarray

class Depth(Sequence):
    """
    The Depth implements the array interface, where every array element, the
    depth, is a numpy.ndarray of a horizontal cut of the volume. As all arrays
    it can be random accessed, iterated over, and read strided. Please note
    that SEG-Y data is laid out trace-by-trace on disk, so accessing horizontal
    cuts (fixed z-coordinates in a cartesian grid) is *very* inefficient.

    This mode works even on unstructured files, because it is not reliant on
    in/crosslines to be sensible. Please note that in the case of unstructured
    depth slicing, the array shape == tracecount.

    Notes
    -----
    .. versionadded:: 1.1

    .. versionchanged:: 1.6
        common list operations (Sequence)

    .. versionchanged:: 1.7.1
       enabled for unstructured files

    .. warning::
        Accessing the file by depth (fixed z-coordinate) is inefficient because
        of poor locality and many reads. If you read more than a handful
        depths, consider using a faster mode.
    """

    def __init__(self, fd):
        super(Depth, self).__init__(len(fd.samples))
        self.filehandle = fd.xfd
        self.dtype = fd.dtype

        if fd.unstructured:
            self.shape = fd.tracecount
            self.offsets = 1
        else:
            self.shape = (len(fd.fast), len(fd.slow))
            self.offsets = len(fd.offsets)

    def __getitem__(self, i):
        """depth[i]

        ith depth, a horizontal cross-section of the file, starting at 0.
        depth[i] returns a numpy.ndarray, and changes to this array will *not*
        be reflected on disk.

        When i is a slice, a generator of numpy.ndarray is returned.

        The depth slices are returned as a fast-by-slow shaped array, i.e. an
        inline sorted file with 10 inlines and 5 crosslines has the shape
        (10,5). If the file is unsorted, the array shape == tracecount.

        Be aware that this interface uses zero-based indices (like traces) and
        *not keys* (like ilines), so you can *not* use the values file.samples
        as indices.

        Parameters
        ----------
        i : int or slice

        Returns
        -------
        depth : numpy.ndarray of dtype or generator of numpy.ndarray of dtype

        Notes
        -----
        .. versionadded:: 1.1

        .. warning::
            The segyio 1.5 and 1.6 series, and 1.7.0, would return the depth_slice in the
            wrong shape for most files. Since segyio 1.7.1, the arrays have the
            correct shape, i.e. fast-by-slow. The underlying data was always
            fast-by-slow, so a numpy array reshape can fix programs using the
            1.5 and 1.6 series.

        Behaves like [] for lists.

        Examples
        --------
        Read a single cut (one sample per trace):

        >>> x = f.depth_slice[199]

        Copy every depth slice into a list:

        >>> l = [numpy.copy(x) for x in depth[:]]

        Every third depth:

        >>> for d in depth[::3]:
        ...     (d * 6).mean()

        Read up to 250:

        >>> for d in depth[:250]:
        ...     d.mean()

        >>> len(ilines), len(xlines)
        (1, 6)
        >>> f.depth_slice[0]
        array([[0.  , 0.01, 0.02, 0.03, 0.04, 0.05]], dtype=float32)
        """

        try:
            i = self.wrapindex(i)
            buf = np.empty(self.shape, dtype=self.dtype)
            return self.filehandle.getdepth(i, buf.size, self.offsets, buf)

        except TypeError:
            try:
                indices = i.indices(len(self))
            except AttributeError:
                msg = 'depth indices must be integers or slices, not {}'
                raise TypeError(msg.format(type(i).__name__))

            def gen():
                x = np.empty(self.shape, dtype=self.dtype)
                y = np.copy(x)

                for j in range(*indices):
                    self.filehandle.getdepth(j, x.size, self.offsets, x)
                    x, y = y, x
                    yield y

            return gen()

    def __setitem__(self, depth, val):
        """depth[i] = val

        Write the ith depth, a horizontal cross-section, of the file, starting
        at 0. It accepts any array_like, but `val` must be at least as big as
        the underlying data slice.

        If `val` is longer than the underlying trace, `val` is essentially truncated.

        Parameters
        ----------
        i   : int or slice
        val : array_like

        Notes
        -----
        .. versionadded:: 1.1

        Behaves like [] for lists.

        Examples
        --------
        Copy a depth:

        >>> depth[4] = other[19]

        Copy consecutive depths, and assign to a sub volume (inject a sub cube
        into the larger volume):

        >>> depth[10:50] = other[:]

        Copy into every other depth from an iterable:

        >>> depth[::2] = other
        """
        if isinstance(depth, slice):
            for i, x in zip(range(*depth.indices(len(self))), val):
                self[i] = x
            return

        val = castarray(val, dtype = self.dtype)
        self.filehandle.putdepth(depth, val.size, self.offsets, val)



from . import Enum

class TraceField(Enum):
    """Trace header field enumerator

    See also
    -------
    segyio.su : Seismic unix aliases for header fields
    """

    TRACE_SEQUENCE_LINE = 1
    TRACE_SEQUENCE_FILE = 5
    FieldRecord = 9
    TraceNumber = 13
    EnergySourcePoint = 17
    CDP = 21
    CDP_TRACE = 25
    TraceIdentificationCode = 29
    NSummedTraces = 31
    NStackedTraces = 33
    DataUse = 35
    offset = 37
    ReceiverGroupElevation = 41
    SourceSurfaceElevation = 45
    SourceDepth = 49
    ReceiverDatumElevation = 53
    SourceDatumElevation = 57
    SourceWaterDepth = 61
    GroupWaterDepth = 65
    ElevationScalar = 69
    SourceGroupScalar = 71
    SourceX = 73
    SourceY = 77
    GroupX = 81
    GroupY = 85
    CoordinateUnits = 89
    WeatheringVelocity = 91
    SubWeatheringVelocity = 93
    SourceUpholeTime = 95
    GroupUpholeTime = 97
    SourceStaticCorrection = 99
    GroupStaticCorrection = 101
    TotalStaticApplied = 103
    LagTimeA = 105
    LagTimeB = 107
    DelayRecordingTime = 109
    MuteTimeStart = 111
    MuteTimeEND = 113
    TRACE_SAMPLE_COUNT = 115
    TRACE_SAMPLE_INTERVAL = 117
    GainType = 119
    InstrumentGainConstant = 121
    InstrumentInitialGain = 123
    Correlated = 125
    SweepFrequencyStart = 127
    SweepFrequencyEnd = 129
    SweepLength = 131
    SweepType = 133
    SweepTraceTaperLengthStart = 135
    SweepTraceTaperLengthEnd = 137
    TaperType = 139
    AliasFilterFrequency = 141
    AliasFilterSlope = 143
    NotchFilterFrequency = 145
    NotchFilterSlope = 147
    LowCutFrequency = 149
    HighCutFrequency = 151
    LowCutSlope = 153
    HighCutSlope = 155
    YearDataRecorded = 157
    DayOfYear = 159
    HourOfDay = 161
    MinuteOfHour = 163
    SecondOfMinute = 165
    TimeBaseCode = 167
    TraceWeightingFactor = 169
    GeophoneGroupNumberRoll1 = 171
    GeophoneGroupNumberFirstTraceOrigField = 173
    GeophoneGroupNumberLastTraceOrigField = 175
    GapSize = 177
    OverTravel = 179
    CDP_X = 181
    CDP_Y = 185
    INLINE_3D = 189
    CROSSLINE_3D = 193
    ShotPoint = 197
    ShotPointScalar = 201
    TraceValueMeasurementUnit = 203
    TransductionConstantMantissa = 205
    TransductionConstantPower = 209
    TransductionUnit = 211
    TraceIdentifier = 213
    ScalarTraceHeader = 215
    SourceType = 217
    SourceEnergyDirectionMantissa = 219
    SourceEnergyDirectionExponent = 223
    SourceMeasurementMantissa = 225
    SourceMeasurementExponent = 229
    SourceMeasurementUnit = 231
    UnassignedInt1 = 233
    UnassignedInt2 = 237

keys = {
    'TRACE_SEQUENCE_LINE'                   : 1,
    'TRACE_SEQUENCE_FILE'                   : 5,
    'FieldRecord'                           : 9,
    'TraceNumber'                           : 13,
    'EnergySourcePoint'                     : 17,
    'CDP'                                   : 21,
    'CDP_TRACE'                             : 25,
    'TraceIdentificationCode'               : 29,
    'NSummedTraces'                         : 31,
    'NStackedTraces'                        : 33,
    'DataUse'                               : 35,
    'offset'                                : 37,
    'ReceiverGroupElevation'                : 41,
    'SourceSurfaceElevation'                : 45,
    'SourceDepth'                           : 49,
    'ReceiverDatumElevation'                : 53,
    'SourceDatumElevation'                  : 57,
    'SourceWaterDepth'                      : 61,
    'GroupWaterDepth'                       : 65,
    'ElevationScalar'                       : 69,
    'SourceGroupScalar'                     : 71,
    'SourceX'                               : 73,
    'SourceY'                               : 77,
    'GroupX'                                : 81,
    'GroupY'                                : 85,
    'CoordinateUnits'                       : 89,
    'WeatheringVelocity'                    : 91,
    'SubWeatheringVelocity'                 : 93,
    'SourceUpholeTime'                      : 95,
    'GroupUpholeTime'                       : 97,
    'SourceStaticCorrection'                : 99,
    'GroupStaticCorrection'                 : 101,
    'TotalStaticApplied'                    : 103,
    'LagTimeA'                              : 105,
    'LagTimeB'                              : 107,
    'DelayRecordingTime'                    : 109,
    'MuteTimeStart'                         : 111,
    'MuteTimeEND'                           : 113,
    'TRACE_SAMPLE_COUNT'                    : 115,
    'TRACE_SAMPLE_INTERVAL'                 : 117,
    'GainType'                              : 119,
    'InstrumentGainConstant'                : 121,
    'InstrumentInitialGain'                 : 123,
    'Correlated'                            : 125,
    'SweepFrequencyStart'                   : 127,
    'SweepFrequencyEnd'                     : 129,
    'SweepLength'                           : 131,
    'SweepType'                             : 133,
    'SweepTraceTaperLengthStart'            : 135,
    'SweepTraceTaperLengthEnd'              : 137,
    'TaperType'                             : 139,
    'AliasFilterFrequency'                  : 141,
    'AliasFilterSlope'                      : 143,
    'NotchFilterFrequency'                  : 145,
    'NotchFilterSlope'                      : 147,
    'LowCutFrequency'                       : 149,
    'HighCutFrequency'                      : 151,
    'LowCutSlope'                           : 153,
    'HighCutSlope'                          : 155,
    'YearDataRecorded'                      : 157,
    'DayOfYear'                             : 159,
    'HourOfDay'                             : 161,
    'MinuteOfHour'                          : 163,
    'SecondOfMinute'                        : 165,
    'TimeBaseCode'                          : 167,
    'TraceWeightingFactor'                  : 169,
    'GeophoneGroupNumberRoll1'              : 171,
    'GeophoneGroupNumberFirstTraceOrigField': 173,
    'GeophoneGroupNumberLastTraceOrigField' : 175,
    'GapSize'                               : 177,
    'OverTravel'                            : 179,
    'CDP_X'                                 : 181,
    'CDP_Y'                                 : 185,
    'INLINE_3D'                             : 189,
    'CROSSLINE_3D'                          : 193,
    'ShotPoint'                             : 197,
    'ShotPointScalar'                       : 201,
    'TraceValueMeasurementUnit'             : 203,
    'TransductionConstantMantissa'          : 205,
    'TransductionConstantPower'             : 209,
    'TransductionUnit'                      : 211,
    'TraceIdentifier'                       : 213,
    'ScalarTraceHeader'                     : 215,
    'SourceType'                            : 217,
    'SourceEnergyDirectionMantissa'         : 219,
    'SourceEnergyDirectionExponent'         : 223,
    'SourceMeasurementMantissa'             : 225,
    'SourceMeasurementExponent'             : 229,
    'SourceMeasurementUnit'                 : 231,
    'UnassignedInt1'                        : 233,
    'UnassignedInt2'                        : 237,
}
